#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
NFS module

Process the NFS layer and return the correct NFS object. The function returns
either a NULL(), CB_NULL, COMPOUND or CB_COMPOUND object.
"""
import nfstest_config as c
from packet.utils import *
from packet.nfs.nfsbase import *
from packet.nfs.nfs3 import NFS3args,NFS3res
from packet.nfs.nfs4 import COMPOUND4args,COMPOUND4res,CB_COMPOUND4args,CB_COMPOUND4res

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

def NFS(rpc, callback):
    """Process the NFS layer and return the correct NFS object"""
    ret = None
    unpack = rpc._pktt.unpack

    if rpc.procedure == 0:
        # NULL object
        if callback:
            ret = CB_NULL()
        else:
            ret = NULL()
    elif rpc.procedure == 1 and ((not callback and rpc.version == 4) or
                           (callback and rpc.version == 1)):
        # NFSv4.x object including callback objects
        if rpc.type == RPC_CALL:
            # RPC call
            if callback:
                ret = CB_COMPOUND4args(unpack)
            else:
                ret = COMPOUND4args(unpack)
        else:
            # RPC reply
            minorversion = None
            pkt_call = rpc._pktt.pkt_call
            if pkt_call is not None and hasattr(pkt_call, "nfs"):
                minorversion = getattr(pkt_call.nfs, "minorversion", None)
            if callback:
                ret = CB_COMPOUND4res(unpack, minorversion)
            else:
                ret = COMPOUND4res(unpack, minorversion)
    elif rpc.version == 3:
        if rpc.type == RPC_CALL:
            # RPC call
            ret = NFS3args(unpack, rpc.procedure)
        else:
            # RPC reply
            ret = NFS3res(unpack, rpc.procedure)

    ret.callback = callback
    return ret



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
NFS Base module

Base class for an NFS object
"""
import nfstest_config as c
from baseobj import BaseObj
import packet.utils as utils
import packet.nfs.nfs4_const as const4

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.3"

# NFSv4 operation priority for displaying purposes
NFSpriority = {
    const4.OP_EXCHANGE_ID          : 95,
    const4.OP_CREATE_SESSION       : 95,
    const4.OP_DESTROY_SESSION      : 95,
    const4.OP_SETCLIENTID          : 95,
    const4.OP_SETCLIENTID_CONFIRM  : 95,
    const4.OP_DESTROY_CLIENTID     : 95,
    const4.OP_RENEW                : 92,
    const4.OP_OPEN                 : 90,
    const4.OP_OPEN_DOWNGRADE       : 90,
    const4.OP_OPENATTR             : 90,
    const4.OP_OPEN_CONFIRM         : 90,
    const4.OP_CLOSE                : 90,
    const4.OP_CREATE               : 90,
    const4.OP_LAYOUTGET            : 85,
    const4.OP_LAYOUTRETURN         : 85,
    const4.OP_LAYOUTCOMMIT         : 85,
    const4.OP_LAYOUTERROR          : 85,
    const4.OP_LAYOUTSTATS          : 85,
    const4.OP_GETDEVICEINFO        : 84,
    const4.OP_GETDEVICELIST        : 83,
    const4.OP_OFFLOAD_STATUS       : 83,
    const4.OP_OFFLOAD_CANCEL       : 83,
    const4.OP_CLONE                : 82,
    const4.OP_COPY                 : 82,
    const4.OP_COPY_NOTIFY          : 82,
    const4.OP_READ                 : 80,
    const4.OP_WRITE                : 80,
    const4.OP_READ_PLUS            : 80,
    const4.OP_WRITE_SAME           : 80,
    const4.OP_COMMIT               : 80,
    const4.OP_SEEK                 : 75,
    const4.OP_LOCK                 : 70,
    const4.OP_LOCKT                : 70,
    const4.OP_LOCKU                : 70,
    const4.OP_RELEASE_LOCKOWNER    : 70,
    const4.OP_ALLOCATE             : 65,
    const4.OP_DEALLOCATE           : 65,
    const4.OP_IO_ADVISE            : 65,
    const4.OP_DELEGRETURN          : 65,
    const4.OP_DELEGPURGE           : 65,
    const4.OP_GET_DIR_DELEGATION   : 63,
    const4.OP_WANT_DELEGATION      : 62,
    const4.OP_LOOKUPP              : 60,
    const4.OP_LOOKUP               : 60,
    const4.OP_READDIR              : 55,
    const4.OP_RENAME               : 50,
    const4.OP_REMOVE               : 50,
    const4.OP_LINK                 : 45,
    const4.OP_SETATTR              : 44,
    const4.OP_READLINK             : 40,
    const4.OP_TEST_STATEID         : 36,
    const4.OP_FREE_STATEID         : 35,
    const4.OP_RECLAIM_COMPLETE     : 34,
    const4.OP_BACKCHANNEL_CTL      : 33,
    const4.OP_BIND_CONN_TO_SESSION : 32,
    const4.OP_PUTROOTFH            : 30,
    const4.OP_PUTPUBFH             : 30,
    const4.OP_ACCESS               : 25,
    const4.OP_SECINFO              : 22,
    const4.OP_SECINFO_NO_NAME      : 22,
    const4.OP_GETATTR              : 20,
    const4.OP_GETFH                : 10,
    const4.OP_SET_SSV              :  7,
    const4.OP_VERIFY               :  0,
    const4.OP_NVERIFY              :  0,
    const4.OP_PUTFH                :  0,
    const4.OP_RESTOREFH            :  0,
    const4.OP_SAVEFH               :  0,
    const4.OP_SEQUENCE             :  0,
    const4.OP_ILLEGAL              :  0,
}

CBpriority = {
    const4.OP_CB_RECALL               : 90,
    const4.OP_CB_LAYOUTRECALL         : 90,
    const4.OP_CB_NOTIFY               : 80,
    const4.OP_CB_OFFLOAD              : 80,
    const4.OP_CB_PUSH_DELEG           : 70,
    const4.OP_CB_RECALL_ANY           : 60,
    const4.OP_CB_RECALLABLE_OBJ_AVAIL : 60,
    const4.OP_CB_RECALL_SLOT          : 50,
    const4.OP_CB_WANTS_CANCELLED      : 40,
    const4.OP_CB_NOTIFY_LOCK          : 30,
    const4.OP_CB_NOTIFY_DEVICEID      : 20,
    const4.OP_CB_GETATTR              : 10,
    const4.OP_CB_SEQUENCE             : 0,
    const4.OP_CB_ILLEGAL              : 0,
}

class NFSbase(utils.RPCload):
    """NFS Base object

       This should only be used as a base class for an NFS object
    """
    def __str__(self):
        """Informal string representation of object"""
        rpc = self._rpc
        rdebug = self.debug_repr()
        if rdebug == 1:
            # String format for verbose level 1
            out = self.rpc_str("NFS")
            if rpc.program >= 0x40000000 and rpc.program < 0x60000000:
                cb_flag  = True
                priority = CBpriority
            else:
                cb_flag  = False
                priority = NFSpriority
            if rpc.procedure == 0:
                # NULL procedure
                out += self.__class__.__name__
                return out
            elif rpc.version == 4 or cb_flag:
                # NFS version 4.x
                if not utils.NFS_mainop:
                    # Display all NFS operation names in the compound
                    oplist = [str(x.op)[3:] for x in self.array]
                    out += "%-25s" % ";".join(oplist)
                if utils.LOAD_body or utils.NFS_mainop:
                    # Order operations by their priority
                    item_list = sorted(self.array, key=lambda x: priority.get(x.op ,0))
                    if utils.NFS_mainop:
                        # Display only the highest priority operation name
                        out += "%-10s" % str(item_list[-1].op)[3:]
                    if utils.LOAD_body:
                        # Get the highest priority operation body to display
                        display_op = None
                        while item_list:
                            item = item_list.pop()
                            if priority.get(item.op, 0) == 0:
                                # Ignore operations with no priority
                                continue
                            itemstr = str(item)
                            if (display_op is None and len(itemstr)) or item.op == display_op:
                                out += " " + itemstr
                                # Check if there is another operation to display
                                display_op = getattr(item, "_opdisp", None)
                                if display_op is None:
                                    break

            if rpc.type and getattr(self, "status", 0) != 0:
                # Display the status of the NFS packet only if it is an error
                out += " %s" % self.status
            return out
        else:
            return BaseObj.__str__(self)

class NULL(NFSbase):
    """NFS NULL object"""
    pass

class CB_NULL(NFSbase):
    """NFS CB_NULL object"""
    pass



#===============================================================================
# Copyright 2015 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
DER decoding module

Decode using ASN.1 DER (Distinguished Encoding Representation)
ASN.1: Abstract Syntax Notation 1

This module does not completely decode all DER data types,
the following is a list of supported data types in this implementation:
    INTEGER,
    BIT_STRING,
    NULL,
    OBJECT_IDENTIFIER,
    GeneralizedTime,
    Strings (OCTET STRING, PrintableString, etc.)
    SEQUENCE OF,
    SEQUENCE,
"""
import re
import time
import struct
import nfstest_config as c
from packet.unpack import Unpack

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2015 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

# DER types
EOC               = 0x00  # End-of-content
BOOLEAN           = 0x01  # Boolean
INTEGER           = 0x02
BIT_STRING        = 0x03  # Bit string
OCTET_STRING      = 0x04  # Octet string
NULL              = 0x05
OBJECT_IDENTIFIER = 0x06  # Object identifier
OBJECT_DESCRIPTOR = 0x07  # Object descriptor
EXTERNAL          = 0x08  # External
REAL              = 0x09  # Floating point number
ENUMERATED        = 0x0a  # Enumerated
EMBEDDED_PDV      = 0x0b  # Embedded PDV
UTF8String        = 0x0c  # UTF8 string
RELATIVE_OID      = 0x0d  # Relative OID
NumericString     = 0x12  # Numeric string
PrintableString   = 0x13  # Printable string
T61String         = 0x14  # T61 string
VideotexString    = 0x15  # Videotex string
IA5String         = 0x16  # IA5 string
UTCTime           = 0x17  # UTC time
GeneralizedTime   = 0x18  # Generalized time
GraphicString     = 0x19  # Graphic string
VisibleString     = 0x1a  # Visible string
GeneralString     = 0x1b  # General string
UniversalString   = 0x1c  # Universal string
CharacterString   = 0x1d  # Character string
BMPString	  = 0x1e  # Unicode string
# DER CONSTRUCTED types
SEQUENCE          = 0x10  # Ordered list of one or more items of different types
SEQUENCE_OF       = 0x10  # Ordered list of one or more items of the same type
SET               = 0x11  # Unordered list of one or more types
SET_OF            = 0x11  # Unordered list of the same types

# ASN.1 form
PRIMITIVE   = 0
CONSTRUCTED = 1

# ASN.1 tagging class
UNIVERSAL   = 0
APPLICATION = 1
CONTEXT     = 2
PRIVATE     = 3

class DERunpack(Unpack):
    """DER unpack object

       Usage:
           from packet.derunpack import DERunpack

           x = DERunpack(buffer)

           # Get the decoded object structure for the stream bytes in buffer
           obj = x.get_item()

           Where obj is of the form:
               obj = {
                   application = {
                       context-tag0 = int|list|dictionary,
                       context-tag1 = int|list|dictionary,
                       ...
                       context-tagN = int|list|dictionary,
                   }
               }

           Example:
             For the following ASN.1 definition:
               TEST ::= [APPLICATION 10] SEQUENCE {
                   id       [0] INTEGER,
                   numbers  [1] SEQUENCE OF INTEGER,
                   data     [2] SEQUENCE {
                       -- NOTE: first tag is [1], not [0]
                       type   [1] INTEGER,
                       value  [2] PrintableString,
                   },
               }

             Using the streamed bytes of the above ASN.1 definition,
             the following is returned by get_item():
               obj = {
                   10 = {              # Application 10
                       0: 53,          # id: context-tag=0, value=53
                       1: [1,2,3],     # numbers: context-tag=1, value=[1,2,3]
                       2: {            # data: context-tag=1, value=structure
                           1: 2,       # id: context-tag=1, value=2
                           2: "test",  # id: context-tag=2, value="test"
                       }
                   }
               }
    """
    def get_tag(self):
        """Get the tag along with the tag class and form or P/C bit

           The first byte(s) of the TLV (Type, Length, Value) is the type
           which has the following format:
             First byte:
               bits 8-7: tag class
               bit 6:    form or P/C (Constructed if bit is set)
               bits 5-1: tag number (0-30)
                         if all bits are 1's (decimal 31) then one or more
                         bytes are required for the tag

             If bits 5-1 are all 1's in the first byte, the tag is given
             in the following bytes:
             Extra byes for tag:
               bit 8:    next byte is part of tag
               bits 7-1: tag bits

           Examples:
             0xa1 (0b10100001): Short form
               tag class: 0b10 = 2 (CONTEXT)
               P/C:       0b0  = 0 (not constructed)

             0x1f8107 (0b000111111000000100000111): Long form
               tag class: 0b00 = 0 (UNIVERSAL -- standard tag)
               P/C:       0b0  = 0 (not constructed)
               tag:   0b11111 = 31 (tag is given in following bytes)
               First extra byte: 0x81 (0b10000001)
                 bit8=1  : there is an extra byte after this
                 bits 7-1: 0b0000001 (0x01 most significant 7 bits of tag)
               Second extra byte: 0x07 (0b00000111)
                 bit8=1  : this is the last byte
                 bits 7-1: 0b0000111 (0x07 least significant 7 bits of tag)
               Tag number: big-endian bits from extra bytes (7 bits each)
                          14 bits: 0x0087 (0x01 << 7 + 0x07) = 135
        """
        tag = self.unpack_uchar()
        self.tclass = tag >> 6
        self.form   = (tag >> 5) & 0x01
        self.tag    = tag & 0x1f
        if tag & 0x1f == 0x1f:
            # Tag is given in the following octets where MSB is set if there
            # is another byte (MSB is 0 for last byte) and the tag number
            # is given by concatenating the 7-bits of all octets
            tag = 0x80
            self.tag = 0
            while tag & 0x80:
                tag = self.unpack_uchar()
                self.tag = (self.tag << 7) + (tag & 0x7f)
        return self.tag

    def get_size(self):
        """Get the size of element (length in TLV)

           Short form: bit8=0, one octet, length given by bits 7-1 (0-127)
           Long form:  bit8=1, 2-127 octet, bits 7-1 give number of length
                       objects

           Example:
             Short form (bit8=0):
               0x0f (0b00001111): length is 0x0f (15)
             Long form (bit8=1 of first byte):
               0x820123 (0b100000100000000100100011):
               length is given by the next 2 bytes (first 7-1 bits 0x02)
               Next two bytes gives the length 0x0123 = 291
        """
        size = self.unpack_uchar()
        if size & 0x80:
            # Long form, get the number of octets for length
            count = size & 0x7f
            # Get length from an unsigned integer of "count" octets
            size = self.der_integer(count, True)
        return size

    def der_integer(self, size=None, unsigned=False):
        """Return an integer given the size of the integer in bytes

           size:
               Number of bytes for the integer, if this option is not given
               the method get_size() is used to get the size of the integer
           unsigned:
               Usually an unsigned integer is encoded with a leading byte
               of all zeros but when decoding data of BIT_STRING type all
               decoded bytes must be unsigned so they can be concatenated
               correctly
        """
        ret = None
        if size is None:
            # If size is not given, get it from the byte stream
            size = self.get_size()
        ret  = 0
        hbit = 0
        for i in range(size):
            byte = self.unpack_uchar()
            if i == 0:
                # Get the most significant bit from the first byte in order
                # to know if this integer is a negative number
                ret = byte
                hbit = byte >> 7
            else:
                ret = (ret << 8) + byte
        if not unsigned and hbit:
            # Convert it to a negative number (two's complement) only if the
            # unsigned option is not given and the most significant bit is set
            ret -= (1<<(8*size))
        return ret

    def der_date(self, size):
        """Return a date time of type GeneralizedTime
           Type GeneralizedTime takes values of the year, month, day, hour,
           minute, second, and second fraction in any of following three forms:

           Local time: "YYYYMMDDHH[MM[SS[.fff]]]"
           Universal time (UTC): "YYYYMMDDHH[MM[SS[.fff]]]Z"
           Difference between local and UTC times" "YYYYMMDDHH[MM[SS[.fff]]]+|-HHMM".

           Where the optional fff is accurate to three decimal places
        """
        data = re.search(r"(\d+)(.(\d+))?(Z?)(([\+\-])(\d\d)(\d\d))?", self.read(size)).groups()
        datestr = data[0]
        if len(datestr) == 14:
            fmt = "%Y%m%d%H%M%S"
        elif len(datestr) == 12:
            fmt = "%Y%m%d%H%M"
        else:
            fmt = "%Y%m%d%H"
        # Local time structure
        ret = time.strptime(datestr, fmt)
        # Convert it to seconds from epoch in current timezone
        utctime = time.mktime(ret)

        tday = 0
        if time.daylight:
            # An hour difference if daylight savings time
            tday = 3600

        if data[3] == "Z":
            # Convert it to UTC including daylight savings time
            utctime -= time.timezone - tday
        elif data[6] is not None and data[7] is not None:
            # Convert it to UTC including daylight savings time
            tz = 3600*eval(data[6]) + 60*eval(data[7])
            if data[5] == "-":
                tz = -tz
            utctime -= tz + time.timezone - tday
        if data[2] is not None:
            # Add the fraction
            slen = len(data[2])
            utctime += int(data[2])/float(10**slen)
        return utctime

    def der_oid(self, size):
        """Return an object identifier (OID)"""
        out = 0
        clist = struct.unpack("!%dB"%size, self.read(size))
        # First byte has the first two nodes
        ret = [str(clist[0]/40), str(clist[0]%40)]
        for item in clist[1:]:
            if item & 0x80:
                # Current node has more bytes
                out = (out << 7) + (item & 0x7f)
            else:
                if out > 0:
                    # This is the last byte for multi-byte node
                    item = (out << 7) + (item & 0x7f)
                ret.append(str(item))
                # Reset multi-byte node
                out = 0
        return ".".join(ret)

    def get_item(self, oid=None):
        """Get item from the byte stream using TLV
           This is a recursive function where the tag and length are decoded
           and then this function is called to get the value if tag is one of
           primitive or non-constructed types.

           Calling this method right after instantiation of the object will
           decode the whole ASN.1 representation
        """
        ret = None
        tagidx = 0
        # Get the Tag
        tag = self.get_tag()
        # Save tag class and P/C
        tclass = self.tclass
        form   = self.form
        # Get the Length
        size = self.get_size()
        if size > len(self):
            # Not enough bytes
            return

        # Get the Value
        if self.tclass in (APPLICATION, CONTEXT) or \
          (self.tclass == UNIVERSAL and self.form == CONSTRUCTED):
            ret = {}
            offset = self.tell()
            while self.tell() - offset < size:
                item = self.get_item()
                if tclass in (APPLICATION, CONTEXT):
                    # Current item (ret) is an Application or Context
                    if tagidx == 1:
                        # Application has more than one item so use implicit
                        # tag numbering
                        ret[tag] = {0:ret[tag], 1:item}
                    elif tagidx > 1:
                        ret[tag][tagidx] = item
                    else:
                        if self.tag == OBJECT_IDENTIFIER and oid is not None and oid == item:
                            ret[tag] = {tagidx:item}
                            break
                        else:
                            ret[tag] = item
                elif self.tclass == CONTEXT:
                    # The item (item) has a Context tag
                    key, value = item.items()[0]
                    ret[key] = value
                else:
                    # Current item (ret) and item have no context tag so this
                    # is a list of simple types (SEQUENCE OF int|string...)
                    # If ret has any items they will be deleted but this should
                    # never happen because all items must have a context tag
                    if isinstance(ret, dict):
                        ret = []
                    ret.append(item)
                if tclass == APPLICATION:
                    tagidx += 1
        elif self.tclass == UNIVERSAL:
            if self.tag == INTEGER:
                ret = self.der_integer(size)
            elif self.tag == BIT_STRING:
                # The first octet in value gives the number of unused bits
                nbits = self.unpack_uchar()
                ret = self.der_integer(size-1, unsigned=True) >> nbits
            elif self.tag == NULL:
                ret = None
            elif self.tag == GeneralizedTime:
                ret = self.der_date(size)
            elif self.tag == OBJECT_IDENTIFIER:
                ret = self.der_oid(size)
            else:
                ret = self.read(size)
        else:
            ret = self.read(size)

        # Restore original tag, tag class and P/C since this method is
        # recursive and a call to get_item() again will modify these values
        self.tag    = tag
        self.form   = form
        self.tclass = tclass
        return ret



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
MacAddr module

Create an object to represent a MAC address. A MAC address is given either
by a series of hexadecimal numbers or using the ":" notation. It provides
a mechanism for comparing this object with a regular string.
"""
import nfstest_config as c

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

class MacAddr(str):
    """MacAddr address object

       Usage:
           from packet.link.macaddr import MacAddr

           mac = MacAddr('E4CE8F589FF4')

       The following expressions are equivalent:
           mac == 'E4CE8F589FF4'
           mac == 'e4ce8f589ff4'
           mac == 'e4:ce:8f:58:9f:f4'
    """
    @staticmethod
    def _convert(mac):
        """Convert string into a persistent representation of a MAC address."""
        if mac != None:
            mac = mac.lower()
            if len(mac) == 12:
                # Add ":" to the string
                t = iter(mac)
                mac = ':'.join(a+b for a,b in zip(t, t))
        return mac

    def __new__(cls, mac):
        """Create new instance by converting input string into a persistent
           representation of a MAC address.
        """
        return super(MacAddr, cls).__new__(cls, MacAddr._convert(mac))

    def __eq__(self, other):
        """Compare two MAC addresses and return True if both are equal."""
        return str(self) == self._convert(other)

    def __ne__(self, other):
        """Compare two MAC addresses and return False if both are equal."""
        return not self.__eq__(other)

if __name__ == '__main__':
    # Self test of module
    mac = MacAddr('E4CE8F589FF4')
    macstr = "%s" % mac
    macrpr = "%r" % mac
    ntests = 6

    tcount = 0
    if mac == 'E4CE8F589FF4':
        tcount += 1
    if mac == 'e4ce8f589ff4':
        tcount += 1
    if mac == 'E4:CE:8F:58:9F:F4':
        tcount += 1
    if mac == 'e4:ce:8f:58:9f:f4':
        tcount += 1
    if macstr == 'e4:ce:8f:58:9f:f4':
        tcount += 1
    if macrpr == "'e4:ce:8f:58:9f:f4'":
        tcount += 1

    if tcount == ntests:
        print "All tests passed!"
        exit(0)
    else:
        print "%d tests failed" % (ntests-tcount)
        exit(1)



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Packet trace module

The Packet trace module is a python module that takes a trace file created
by tcpdump and unpacks the contents of each packet. You can decode one packet
at a time, or do a search for specific packets. The main difference between
these modules and other tools used to decode trace files is that you can use
this module to completely automate your tests.

How does it work? It opens the trace file and reads one record at a time
keeping track where each record starts. This way, very large trace files
can be opened without having to wait for the file to load and avoid loading
the whole file into memory.

Packet layers supported:
    - ETHERNET II (RFC 894)
    - IP layer (supports IPv4 and IPv6)
    - UDP layer
    - TCP layer
    - RPC layer
    - NFS v4.0
    - NFS v4.1 including pNFS file layouts
    - NFS v4.2
    - PORTMAP v2
    - MOUNT v3
    - NLM v4
"""
import os
import re
import sys
import gzip
import time
import fcntl
import token
import struct
import parser
import symbol
import termios
from formatstr import *
import nfstest_config as c
from baseobj import BaseObj
from packet.unpack import Unpack
from packet.record import Record
from packet.pkt import Pkt, PKT_layers
from packet.link.ethernet import ETHERNET

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "2.3"

BaseObj.debug_map(0x100000000, 'pkt1', "PKT1: ")
BaseObj.debug_map(0x200000000, 'pkt2', "PKT2: ")
BaseObj.debug_map(0x400000000, 'pkt3', "PKT3: ")
BaseObj.debug_map(0x800000000, 'pkt4', "PKT4: ")
BaseObj.debug_map(0xF00000000, 'pktt', "PKTT: ")

# Map of tokens
_token_map = dict(token.tok_name.items() + symbol.sym_name.items())
# Map of items not in the array of the compound
_nfsopmap = {'status': 1, 'tag': 1}
# Match function map
_match_func_map = dict(zip(PKT_layers,["self._match_%s"%x for x in PKT_layers]))

# Read size -- the amount of data read at a time from the file
# The read ahead buffer actual size is always >= 2*READ_SIZE
READ_SIZE = 64*1024

# Show progress if stderr is a tty and stdout is not
SHOWPROG = os.isatty(2) and not os.isatty(1)

class Header(BaseObj):
    # Class attributes
    _attrlist = ("major", "minor", "zone_offset", "accuracy",
                 "dump_length", "link_type")

    def __init__(self, pktt):
        ulist = struct.unpack(pktt.header_fmt, pktt._read(20))
        self.major       = ulist[0]
        self.minor       = ulist[1]
        self.zone_offset = ulist[2]
        self.accuracy    = ulist[3]
        self.dump_length = ulist[4]
        self.link_type   = ulist[5]

class Pktt(BaseObj, Unpack):
    """Packet trace object

       Usage:
           from packet.pktt import Pktt

           x = Pktt("/traces/tracefile.cap")

           # Iterate over all packets found in the trace file
           for pkt in x:
               print pkt
    """
    def __init__(self, tfile, live=False, state=True):
        """Constructor

           Initialize object's private data, note that this will not check the
           file for existence nor will open the file to verify if it is a valid
           tcpdump file. The tcpdump trace file will be opened the first time a
           packet is retrieved.

           tracefile:
               Name of tcpdump trace file or a list of trace file names
               (little or big endian format)
           live:
               If set to True, methods will not return if encountered <EOF>,
               they will keep on trying until more data is available in the
               file. This is useful when running tcpdump in parallel,
               especially when tcpdump is run with the '-C' option, in which
               case when <EOF> is encountered the next trace file created by
               tcpdump will be opened and the object will be re-initialized,
               all private data referencing the previous file is lost.
        """
        self.tfile   = tfile  # Current trace file name
        self.bfile   = tfile  # Base trace file name
        self.live    = live   # Set to True if dealing with a live tcpdump file
        self.offset  = 0      # Current file offset
        self.boffset = 0      # File offset of current packet
        self.ioffset = 0      # File offset of first packet
        self.index   = 0      # Current packet index
        self.frame   = 1      # Current frame number
        self.mindex  = 0      # Maximum packet index for current trace file
        self.findex  = 0      # Current tcpdump file index (used with self.live)
        self.pindex  = 0      # Current packet index (for pktlist)
        self.pktlist = None   # Match from this packet list instead
        self.fh      = None   # Current file handle
        self.eof     = False  # End of file marker for current packet trace
        self.serial  = False  # Processing trace files serially
        self.pkt     = None   # Current packet
        self.pkt_call  = None # The current packet call if self.pkt is a reply
        self.pktt_list = []   # List of Pktt objects created
        self.tfiles    = []   # List of packet trace files
        self.rdbuffer  = ""   # Read buffer
        self.rdoffset  = 0    # Read buffer offset
        self.filesize  = 0    # Size of packet trace file
        self.prevprog  = -1.0 # Previous progress percentage
        self.prevtime  = 0.0  # Previous segment time
        self.prevdone  = -1   # Previous progress bar units done so far
        self.prevoff   = 0    # Previous offset
        self.showprog  = 0    # If this is true the progress will be displayed
        self.progdone  = 0    # Display last progress only once
        self.timestart = time.time() # Time reference base
        self.reply_matched = False   # Matching a reply

        # TCP stream map: to keep track of the different TCP streams within
        # the trace file -- used to deal with RPC packets spanning multiple
        # TCP packets or to handle a TCP packet having multiple RPC packets
        self._tcp_stream_map = {}

        # RPC xid map: to keep track of packet calls
        self._rpc_xid_map = {}
        # List of outstanding xids to match
        self._match_xid_list = []

        # Process tfile argument
        if isinstance(tfile, list):
            # The argument tfile is given as a list of packet trace files
            self.tfiles = tfile
            if len(self.tfiles) == 1:
                # Only one file is given
                self.tfile = self.tfiles[0]
            else:
                # Create all packet trace objects
                for tfile in self.tfiles:
                    self.pktt_list.append(Pktt(tfile))

    def __del__(self):
        """Destructor

           Gracefully close the tcpdump trace file if it is opened.
        """
        if self.fh:
            self.fh.close()

    def __iter__(self):
        """Make this object iterable."""
        return self

    def __contains__(self, expr):
        """Implement membership test operator.
           Return true if expr matches a packet in the trace file,
           false otherwise.

           The packet is also stored in the object attribute pkt.

           Examples:
               # Find the next READ request
               if ("NFS.argop == 25" in x):
                   print x.pkt.nfs

           See match() method for more information
        """
        pkt = self.match(expr)
        return (pkt is not None)

    def __getitem__(self, index):
        """Get the packet from the trace file given by the index
           or raise IndexError.

           The packet is also stored in the object attribute pkt.

           Examples:
               pkt = x[index]
        """
        self.dprint('PKT4', ">>> __getitem__(%d)" % index)
        if index < 0:
            # No negative index is allowed
            raise IndexError

        try:
            if index == self.pkt.record.index:
                # The requested packet is in memory, just return it
                return self.pkt
        except:
            pass

        if index < self.index:
            # Reset the current packet index and offset
            # The index is less than the current packet offset so position
            # the file pointer to the offset of the packet given by index
            self.rewind(index)

        # Move to the packet specified by the index
        pkt = None
        while self.index <= index:
            try:
                pkt = self.next()
            except:
                break

        if pkt is None:
            raise IndexError
        return pkt

    def next(self):
        """Get the next packet from the trace file or raise StopIteration.

           The packet is also stored in the object attribute pkt.

           Examples:
               # Iterate over all packets found in the trace file using
               # the iterable properties of the object
               for pkt in x:
                   print pkt

               # Iterate over all packets found in the trace file using it
               # as a method and using the object variable as the packet
               # Must use the try statement to catch StopIteration exception
               try:
                   while (x.next()):
                       print x.pkt
               except StopIteration:
                   pass

               # Iterate over all packets found in the trace file using it
               # as a method and using the return value as the packet
               # Must use the try statement to catch StopIteration exception
               while True:
                   try:
                       print x.next()
                   except StopIteration:
                       break

           NOTE:
               Supports only single active iteration
        """
        self.dprint('PKT4', ">>> %d: next()" % self.index)
        # Initialize next packet
        self.pkt = Pkt()

        if len(self.pktt_list) > 1:
            # Dealing with multiple trace files
            minsecs  = None
            pktt_obj = None
            for obj in self.pktt_list:
                if obj.pkt is None:
                    # Get first packet for this packet trace object
                    try:
                        obj.next()
                    except StopIteration:
                        obj.mindex = self.index
                if obj.eof:
                    continue
                if minsecs is None or obj.pkt.record.secs < minsecs:
                    minsecs = obj.pkt.record.secs
                    pktt_obj = obj
            if self.filesize == 0:
                # Calculate total bytes to process
                for obj in self.pktt_list:
                    self.filesize += obj.filesize
            if pktt_obj is None:
                # All packet trace files have been processed
                self.offset = self.filesize
                self.show_progress(True)
                raise StopIteration
            elif len(self._tcp_stream_map):
                # This packet trace file should be processed serially
                # Have all state transferred to next packet object
                pktt_obj.rewind()
                pktt_obj._tcp_stream_map = self._tcp_stream_map
                pktt_obj._rpc_xid_map    = self._rpc_xid_map
                self._tcp_stream_map = {}
                self._rpc_xid_map    = {}
                pktt_obj.next()

            # Overwrite attributes seen by the caller with the attributes
            # from the current packet trace object
            self.pkt = pktt_obj.pkt
            self.pkt_call = pktt_obj.pkt_call
            self.tfile = pktt_obj.tfile
            self.pkt.record.index = self.index  # Use a cumulative index
            self.offset += pktt_obj.offset - pktt_obj.boffset

            try:
                # Get next packet for this packet trace object
                pktt_obj.next()
            except StopIteration:
                # Set maximum packet index for this packet trace object to
                # be used by rewind to select the proper packet trace object
                pktt_obj.mindex = self.index
                # Check if objects should be serially processed
                pktt_obj.serial = False
                for obj in self.pktt_list:
                    if not obj.eof:
                        if obj.index > 1:
                            pktt_obj.serial = False
                            break
                        elif obj.index == 1:
                            pktt_obj.serial = True
                if pktt_obj.serial:
                    # Save current state
                    self._tcp_stream_map = pktt_obj._tcp_stream_map
                    self._rpc_xid_map    = pktt_obj._rpc_xid_map

            self.show_progress()

            # Increment cumulative packet index
            self.index += 1
            return self.pkt

        if self.boffset != self.offset:
            # Frame number is one for every record header on the pcap trace
            # On the other hand self.index is the packet number. Since there
            # could be multiple packets on a single frame self.index could
            # be larger the self.frame except that self.index start at 0
            # while self.frame starts at 1.
            # The frame number can be used to match packets with other tools
            # like wireshark
            self.frame += 1

        # Save file offset for this packet
        self.boffset = self.offset

        # Get record header
        data = self._read(16)
        if len(data) < 16:
            self.eof = True
            self.offset = self.filesize
            self.show_progress(True)
            raise StopIteration
        # Decode record header
        record = Record(self, data)

        # Get record data and create Unpack object
        self.unpack = Unpack(self._read(record.length_inc))
        if self.unpack.size() < record.length_inc:
            # Record has been truncated, stop iteration
            self.eof = True
            self.offset = self.filesize
            self.show_progress(True)
            raise StopIteration

        if self.header.link_type == 1:
            # Decode ethernet layer
            ETHERNET(self)
        else:
            # Unknown link layer
            record.data = self.unpack.getbytes()

        self.show_progress()

        # Increment packet index
        self.index += 1

        return self.pkt

    def rewind(self, index=0):
        """Rewind the trace file by setting the file pointer to the start of
           the given packet index. Returns False if unable to rewind the file,
           e.g., when the given index is greater than the maximum number
           of packets processed so far.
        """
        self.dprint('PKT1', ">>> rewind(%d)" % index)
        if self.pktlist is not None:
            self.pindex = index
            return True
        if index >= 0 and index < self.index:
            if len(self.pktt_list) > 1:
                # Dealing with multiple trace files
                self.index = 0
                for obj in self.pktt_list:
                    if not obj.eof or index <= obj.mindex:
                        obj.rewind()
                        try:
                            obj.next()
                        except StopIteration:
                            pass
                    elif obj.serial and index > obj.mindex:
                        self.index = obj.mindex + 1
            else:
                # Reset the current packet index and offset to the first packet
                self.offset = self.ioffset
                self.index  = 0
                self.eof    = False

                # Position the file pointer to the offset of the first packet
                self.seek(self.ioffset)

                # Clear state
                self._tcp_stream_map = {}
                self._rpc_xid_map    = {}

            # Move to the packet before the specified by the index so the
            # next packet fetched will be the one given by index
            while self.index < index:
                try:
                    pkt = self.next()
                except:
                    break

            # Rewind succeeded
            return True
        return False

    def seek(self, offset, whence=os.SEEK_SET, hard=False):
        """Position the read offset correctly
           If new position is outside the current read buffer then clear the
           buffer so a new chunk of data will be read from the file instead
        """
        soffset = self.fh.tell() - len(self.rdbuffer)
        if hard or offset < soffset or whence != os.SEEK_SET:
            # Seek is before the read buffer, do the actual seek
            self.rdbuffer = ""
            self.rdoffset = 0
            self.fh.seek(offset, whence)
            self.offset = self.fh.tell()
        else:
            # Seek is not before the read buffer
            self.rdoffset = offset - soffset
            self.offset = offset

    def _getfh(self):
        """Get the filehandle of the trace file, open file if necessary."""
        if self.fh == None:
            # Check size of file
            fstat = os.stat(self.tfile)
            if fstat.st_size == 0:
                raise Exception("Packet trace file is empty")

            # Open trace file
            self.fh = open(self.tfile, 'rb')
            self.filesize = fstat.st_size

            iszip = False
            self.header_fmt = None
            while self.header_fmt is None:
                # Initialize offset
                self.offset = 0

                # Get file identifier
                try:
                    self.ident = self._read(4)
                except:
                    self.ident = ""

                if self.ident == '\324\303\262\241':
                    # Little endian
                    self.header_fmt = '<HHIIII'
                    self.header_rec = '<IIII'
                elif self.ident == '\241\262\303\324':
                    # Big endian
                    self.header_fmt = '>HHIIII'
                    self.header_rec = '>IIII'
                else:
                    if iszip:
                        raise Exception('Not a tcpdump file')
                    iszip = True
                    # Get the size of the uncompressed file, this only works
                    # for uncompressed files less than 4GB
                    self.fh.seek(-4, os.SEEK_END)
                    self.filesize = struct.unpack("<I", self.fh.read(4))[0]
                    # Do a hard seek -- clear read ahead buffer
                    self.seek(0, hard=True)
                    # Try if this is a gzip compress file
                    self.fh = gzip.GzipFile(fileobj=self.fh)

            # Get header information
            self.header = Header(self)

            # Initialize packet number
            self.index   = 0
            self.tstart  = None
            self.ioffset = self.offset

        return self.fh

    def _read(self, count):
        """Wrapper for read in order to increment the object's offset. It also
           takes care of <EOF> when 'live' option is set which keeps on trying
           to read and switching files when needed.
        """
        # Open packet trace if needed
        self._getfh()
        while True:
            # Get the number of bytes specified
            rdsize = len(self.rdbuffer) - self.rdoffset
            if count > rdsize:
                # Not all bytes needed are in the read buffer
                if self.rdoffset > READ_SIZE:
                    # If the read offset is on the second half of the
                    # 2*READ_SIZE buffer discard the first bytes so the
                    # new read offset is right at the middle of the buffer
                    # This is done in case there is a seek behind the current
                    # offset so data is not read from the file again
                    self.rdbuffer = self.rdbuffer[self.rdoffset-READ_SIZE:]
                    self.rdoffset = READ_SIZE
                # Read next chunk from file
                self.rdbuffer += self.fh.read(max(count, READ_SIZE))
            # Get the bytes requested and increment read offset accordingly
            data = self.rdbuffer[self.rdoffset:self.rdoffset+count]
            self.rdoffset += count

            ldata = len(data)
            if self.live and ldata != count:
                # Not all data was read (<EOF>)
                tracefile = "%s%d" % (self.bfile, self.findex+1)
                # Check if next trace file exists
                if os.path.isfile(tracefile):
                    # Save information that keeps track of the next trace file
                    basefile = self.bfile
                    findex = self.findex + 1
                    # Re-initialize the object
                    self.__del__()
                    self.__init__(tracefile, live=self.live)
                    # Overwrite next trace file info
                    self.bfile = basefile
                    self.findex = findex
                # Re-position file pointer to last known offset
                self.seek(self.offset)
                time.sleep(1)
            else:
                break

        # Increment object's offset by the amount of data read
        self.offset += ldata
        return data

    def _split_match(self, uargs):
        """Split match arguments and return a tuple (lhs, opr, rhs)
           where lhs is the left hand side of the given argument expression,
           opr is the operation and rhs is the right hand side of the given
           argument expression:

               <lhs> <opr> <rhs>

           Valid opr values are: ==, !=, <, >, <=, >=, in
        """
        m = re.search(r"([^!=<>]+)\s*([!=<>]+|in)\s*(.*)", uargs)
        lhs = m.group(1).rstrip()
        opr = m.group(2)
        rhs = m.group(3)
        return (lhs, opr, rhs)

    def _process_match(self, obj, lhs, opr, rhs):
        """Process "regex" and 'in' operator on match expression.
           Regular expression is given as re('regex') and converted to a
           proper regex re.search('regex', data), where data is the object
           compose of obj and lhs|rhs depending on opr. The argument obj
           is an object prefix.

           If opr is a comparison operation (==, !=, etc.), both obj and lhs
           will be the actual LHS and rhs will be the actual RHS.
           If opr is 'in', lhs will be the actual LHS and both obj and rhs
           will be the actual RHS.

           Return the processed match expression.

           Examples:
               # Regular expression processing
               expr = x._process_match('self.pkt.ip.', 'src', '==', "re(r'192\.*')")

               Returns the following expression ready to be evaluated:
               expr = "re.search(r'192\.*', str(self.pkt,ip.src))"

               # Object prefix processing
               expr = x._process_match('item.', 'argop', '==', '25')

               Returns the following expression ready to be evaluated:
               expr = "item.argop==25"

               # Membership (in) processing
               expr = x._process_match('item.', '62', 'in', 'attributes')

               Returns the following expression ready to be evaluated:
               expr = "62 in item.attributes"
        """
        func = None
        if opr != 'in':
            regex = re.search(r"(\w+)\((.*)\)", lhs)
            if regex:
                lhs = regex.group(2)
                func = regex.group(1)

        if rhs[:3] == 're(':
            # Regular expression, it must be in rhs
            rhs = "re.search" + rhs[2:]
            if opr == "!=":
                rhs = "not " + rhs
            LHS = rhs[:-1] + ", str(" + obj + lhs +  "))"
            RHS = ""
            opr = ""
        elif opr == 'in':
            opr = " in "
            if self.inlhs:
                LHS = obj + lhs
                RHS = rhs
            else:
                LHS = lhs
                RHS = obj + rhs
        else:
            LHS = obj + lhs
            RHS = rhs

        if func is not None:
            LHS = "%s(%s)" % (func, LHS)

        return LHS + opr + RHS

    def _match(self, layer, uargs):
        """Default match function."""
        if not hasattr(self.pkt, layer):
            return False

        if layer == "nfs":
            # Use special matching function for NFS
            texpr = self.match_nfs(uargs)
        else:
            # Use general match
            obj = "self.pkt.%s." % layer.lower()
            lhs, opr, rhs = self._split_match(uargs)
            expr = self._process_match(obj, lhs, opr, rhs)
            texpr = eval(expr)
        self.dprint('PKT2', "    %d: match_%s(%s) -> %r" % (self.pkt.record.index, layer, uargs, texpr))
        return texpr

    def get_index(self):
        """Get current packet index"""
        if self.pktlist is None:
            return self.index
        else:
            return self.pindex

    def set_pktlist(self, pktlist=None):
        """Set the current packet list for buffered matching in which the
           match method will only use this list instead of getting the next
           packet from the packet trace file.
           This could be used when there is a lot of matching going back
           and forth but only on a particular set of packets.
           See the match() method for an example of buffered matching.
        """
        self.pindex  = 0
        self.pktlist = pktlist

    def clear_xid_list(self):
        """Clear list of outstanding xids"""
        self._match_xid_list = []

    def _match_nfs(self, uargs):
        """Match NFS values on current packet."""
        array = None
        isarg = True
        lhs, opr, rhs = self._split_match(uargs)

        if self.pkt.rpc.version == 3 or _nfsopmap.get(lhs):
            try:
                # Top level NFSv4 packet info or NFSv3 packet
                expr = self._process_match("self.pkt.nfs.", lhs, opr, rhs)
                if eval(expr):
                    # Set NFSop and NFSidx
                    self.pkt.NFSop = self.pkt.nfs
                    self.pkt.NFSidx = 0
                    return True
                return False
            except Exception:
                return False

        idx = 0
        obj_prefix = "item."
        for item in self.pkt.nfs.array:
            try:
                # Get expression to eval
                expr = self._process_match(obj_prefix, lhs, opr, rhs)
                if eval(expr):
                    self.pkt.NFSop = item
                    self.pkt.NFSidx = idx
                    return True
            except Exception:
                # Continue searching
                pass
            idx += 1
        return False

    def match_nfs(self, uargs):
        """Match NFS values on current packet.

           In NFSv4, there is a single compound procedure with multiple
           operations, matching becomes a little bit tricky in order to make
           the matching expression easy to use. The NFS object's name space
           gets converted into a flat name space for the sole purpose of
           matching. In other words, all operation objects in array are
           treated as being part of the NFS object's top level attributes.

           Consider the following NFS object:
               nfsobj = COMPOUND4res(
                   status=NFS4_OK,
                   tag='NFSv4_tag',
                   array = [
                       nfs_resop4(
                           resop=OP_SEQUENCE,
                           opsequence=SEQUENCE4res(
                               status=NFS4_OK,
                               resok=SEQUENCE4resok(
                                   sessionid='sessionid',
                                   sequenceid=29,
                                   slotid=0,
                                   highest_slotid=179,
                                   target_highest_slotid=179,
                                   status_flags=0,
                               ),
                           ),
                       ),
                       nfs_resop4(
                           resop=OP_PUTFH,
                           opputfh = PUTFH4res(
                               status=NFS4_OK,
                           ),
                       ),
                       ...
                   ]
               ),

           The result for operation PUTFH is the second in the list:
               putfh = nfsobj.array[1]

           From this putfh object the status operation can be accessed as:
               status = putfh.opputfh.status

           or simply as (this is how the NFS object works):
               status = putfh.status

           In this example, the following match expression 'NFS.status == 0'
           could match the top level status of the compound (nfsobj.status)
           or the putfh status (nfsobj.array[1].status)

           The following match expression 'NFS.sequenceid == 25' will also
           match this packet as well, even though the actual expression should
           be 'nfsobj.array[0].opsequence.resok.sequenceid == 25' or
           simply 'nfsobj.array[0].sequenceid == 25'.

           This approach makes the match expressions simpler at the expense of
           having some ambiguities on where the actual match occurred. If a
           match is desired on a specific operation, a more qualified name can
           be given. In the above example, in order to match the status of the
           PUTFH operation the match expression 'NFS.opputfh.status == 0' can
           be used. On the other hand, consider a compound having multiple
           PUTFH results the above match expression will always match the first
           occurrence of PUTFH where the status is 0. There is no way to tell
           the match engine to match the second or Nth occurrence of an
           operation.
        """
        texpr = self._match_nfs(uargs)
        self.dprint('PKT2', "    %d: match_nfs(%s) -> %r" % (self.pkt.record.index, uargs, texpr))
        return texpr

    def _convert_match(self, ast):
        """Convert a parser list match expression into their corresponding
           function calls.

           Example:
               expr = "TCP.flags.ACK == 1 and NFS.argop == 50"
               st = parser.expr(expr)
               ast = parser.st2list(st)
               data =  self._convert_match(ast)

               Returns:
               data = "(self._match('tcp','flags.ACK==1'))and(self._match('nfs','argop==50'))"
        """
        ret = ''
        isin = False
        if not isinstance(ast, list):
            if ast.lower() in _match_func_map:
                # Replace name by its corresponding function name
                return _match_func_map[ast.lower()]
            return ast
        if len(ast) == 2:
            return self._convert_match(ast[1])

        for a in ast[1:]:
            data = self._convert_match(a)
            if data == 'in':
                data = ' in '
                isin = True
                if ret[:5] == "self.":
                    # LHS in the 'in' operator is a packet object
                    self.inlhs = True
                else:
                    # LHS in the 'in' operator is a constant value
                    self.inlhs = False
            ret += data

        if _token_map[ast[0]] == "comparison":
            # Comparison
            if isin:
                regex = re.search(r'(.*)(self\._match)_(\w+)\.(.*)', ret)
                data  = regex.groups()
                func  = data[1]
                layer = data[2]
                uargs = data[0] + data[3]
            else:
                regex = re.search(r"^((\w+)\()?(self\._match)_(\w+)\.(.*)", ret)
                data  = regex.groups()
                func  = data[2]
                layer = data[3]
                if data[0] is None:
                    uargs = data[4]
                else:
                    uargs = data[0] + data[4]
            # Escape all single quotes since the whole string will be quoted
            uargs = re.sub(r"'", "\\'", uargs)
            ret = "(%s('%s','%s'))" % (func, layer, uargs)

        return ret

    def match(self, expr, maxindex=None, rewind=True, reply=False):
        """Return the packet that matches the given expression, also the packet
           index points to the next packet after the matched packet.
           Returns None if packet is not found and the packet index points
           to the packet at the beginning of the search.

           expr:
               String of expressions to be evaluated
           maxindex:
               The match fails if packet index hits this limit
           rewind:
               Rewind to index where matching started if match fails
           reply:
               Match RPC replies of previously matched calls as well

           Examples:
               # Find the packet with both the ACK and SYN TCP flags set to 1
               pkt = x.match("TCP.flags.ACK == 1 and TCP.flags.SYN == 1")

               # Find the next NFS EXCHANGE_ID request
               pkt = x.match("NFS.argop == 42")

               # Find the next NFS EXCHANGE_ID or CREATE_SESSION request
               pkt = x.match("NFS.argop in [42,43]")

               # Find the next NFS OPEN request or reply
               pkt = x.match("NFS.op == 18")

               # Find all packets coming from subnet 192.168.1.0/24 using
               # a regular expression
               while x.match(r"IP.src == re('192\.168\.1\.\d*')"):
                   print x.pkt.tcp

               # Find packet having a GETATTR asking for FATTR4_FS_LAYOUT_TYPES(bit 62)
               pkt_call = x.match("NFS.attr_request & 0x4000000000000000L != 0")
               if pkt_call:
                   # Find GETATTR reply
                   xid = pkt_call.rpc.xid
                   # Find reply where the number 62 is in the array NFS.attributes
                   pkt_reply = x.match("RPC.xid == %d and 62 in NFS.attributes" % xid)

               # Find the next WRITE request
               pkt = x.match("NFS.argop == 38")
               if pkt:
                   print pkt.nfs

               # Same as above, but using membership test operator instead
               if ("NFS.argop == 38" in x):
                   print x.pkt.nfs

               # Get a list of all OPEN and CLOSE packets then use buffered
               # matching to process each OPEN and its corresponding CLOSE
               # at a time including both requests and replies
               pktlist = []
               while x.match("NFS.op in [4,18]"):
                   pktlist.append(x.pkt)
               # Enable buffered matching
               x.set_pktlist(pktlist)
               while x.match("NFS.argop == 18"): # Find OPEN request
                   print x.pkt
                   index = x.get_index()
                   # Find OPEN reply
                   x.match("RPC.xid == %d and NFS.resop == 18" % x.pkt.rpc.xid)
                   print x.pkt
                   # Find corresponding CLOSE request
                   stid = x.escape(x.pkt.NFSop.stateid.other)
                   x.match("NFS.argop == 4 and NFS.stateid == '%s'" % stid)
                   print x.pkt
                   # Find CLOSE reply
                   x.match("RPC.xid == %d and NFS.resop == 4" % x.pkt.rpc.xid)
                   print x.pkt
                   # Rewind to right after the OPEN request
                   x.rewind(index)
               # Disable buffered matching
               x.set_pktlist()

           See also:
               match_ethernet(), match_ip(), match_tcp(), match_rpc(), match_nfs()
        """
        # Parse match expression
        st = parser.expr(expr)
        smap = parser.st2list(st)
        pdata = self._convert_match(smap)
        self.dprint('PKT1', ">>> %d: match(%s)" % (self.index, expr))
        self.reply_matched = False
        if self.pktlist is None:
            pkt_list   = self
            save_index = self.index
        else:
            pkt_list   = self.pktlist
            save_index = self.pindex

        # Search one packet at a time
        for pkt in pkt_list:
            if maxindex and self.index > maxindex:
                # Hit maxindex limit
                break
            if self.pktlist is not None:
                if pkt.record.index < self.pindex:
                    continue
                else:
                    self.pindex = pkt.record.index + 1
                    self.pkt = pkt
            try:
                if reply and pkt == "rpc" and pkt.rpc.type == 1 and pkt.rpc.xid in self._match_xid_list:
                    self.dprint('PKT1', ">>> %d: match() -> True: reply" % pkt.record.index)
                    self._match_xid_list.remove(pkt.rpc.xid)
                    self.reply_matched = True
                    return pkt
                if eval(pdata):
                    # Return matched packet
                    self.dprint('PKT1', ">>> %d: match() -> True" % pkt.record.index)
                    if reply and pkt == "rpc" and pkt.rpc.type == 0:
                        # Save xid of matched call
                        self._match_xid_list.append(pkt.rpc.xid)
                    return pkt
            except Exception:
                pass

        if rewind:
            # No packet matched, re-position the file pointer back to where
            # the search started
            if self.pktlist is None:
                self.rewind(save_index)
            else:
                self.pindex = save_index
        self.pkt = None
        self.dprint('PKT1', ">>> match() -> False")
        return None

    def show_progress(self, done=False):
        """Display progress bar if enabled and if running on correct terminal"""
        if SHOWPROG and self.showprog and (done or self.index % 500 == 0) \
          and (os.getpgrp() == os.tcgetpgrp(sys.stderr.fileno())):
            rows, columns = struct.unpack('hh', fcntl.ioctl(2, termios.TIOCGWINSZ, "1234"))
            if columns < 100:
                sps = 30
            else:
                # Terminal is wide enough, include bytes/sec
                sps = 42
            # Progress bar length
            wlen = int(columns) - len(str_units(self.filesize)) - sps
            # Number of bytes per progress bar unit
            xunit = float(self.filesize)/wlen
            # Progress bar units done so far
            xdone = int(self.offset/xunit)
            xtime = time.time()
            progress = 100.0*self.offset/self.filesize

            # Display progress only if there is some change in progress
            if (done and not self.progdone) or (self.prevdone != xdone or \
               int(self.prevtime) != int(xtime) or \
               round(self.prevprog) != round(progress)):
                if done:
                    # Do not display progress again when done=True
                    self.progdone = 1
                otime  = xtime - self.timestart # Overall time
                tdelta = xtime - self.prevtime  # Segment time
                self.prevprog = progress
                self.prevdone = xdone
                self.prevtime = xtime
                # Number of progress bar units for completion
                slen = wlen - xdone
                if done:
                    # Overall average bytes/sec
                    bps = self.offset / otime
                else:
                    # Segment average bytes/sec
                    bps = (self.offset - self.prevoff) / tdelta
                self.prevoff = self.offset
                # Progress bar has both foreground and background colors
                # as green and in case the terminal does not support colors
                # then a "=" is displayed instead instead of a green block
                pbar = " [\033[32m\033[42m%s\033[m%s] " % ("="*xdone, " "*slen)
                # Add progress percentage and how many bytes have been
                # processed so far relative to the total number of bytes
                pbar += "%5.1f%% %9s/%s" % (progress, str_units(self.offset), str_units(self.filesize))
                if columns < 100:
                    sys.stderr.write("%s %-6s\r" % (pbar, str_time(otime)))
                else:
                    # Terminal is wide enough, include bytes/sec
                    sys.stderr.write("%s %9s/s %-6s\r" % (pbar, str_units(bps), str_time(otime)))
                if done:
                    sys.stderr.write("\n")

    @staticmethod
    def escape(data):
        """Escape special characters.

           Examples:
               # Call as an instance
               escaped_data = x.escape(data)

               # Call as a class
               escaped_data = Pktt.escape(data)
        """
        # repr() can escape or not a single quote depending if a double
        # quote is present, just make sure both quotes are escaped correctly
        rdata = repr(data)
        if rdata[0] == '"':
            # Double quotes are escaped
            dquote = r'x22'
            squote = r'\x27'
        else:
            # Single quotes are escaped
            dquote = r'\x22'
            squote = r'x27'
        # Replace all double quotes to its corresponding hex value
        data = re.sub(r'"', dquote, rdata[1:-1])
        # Replace all single quotes to its corresponding hex value
        data = re.sub(r"'", squote, data)
        # Escape all backslashes
        data = re.sub(r'\\', r'\\\\', data)
        return data

    @staticmethod
    def ip_tcp_src_expr(ipaddr, port):
        """Return a match expression to find a packet coming from ipaddr:port.

           Examples:
               # Call as an instance
               expr = x.ip_tcp_src_expr('192.168.1.50', 2049)

               # Call as a class
               expr = Pktt.ip_tcp_src_expr('192.168.1.50', 2049)

               # Returns "IP.src == '192.168.1.50' and TCP.src_port == 2049"
               # Expression ready for x.match()
               pkt = x.match(expr)
        """
        return "IP.src == '%s' and TCP.src_port == %d" % (ipaddr, port)

    @staticmethod
    def ip_tcp_dst_expr(ipaddr, port):
        """Return a match expression to find a packet going to ipaddr:port.

           Examples:
               # Call as an instance
               expr = x.ip_tcp_dst_expr('192.168.1.50', 2049)

               # Call as a class
               expr = Pktt.ip_tcp_dst_expr('192.168.1.50', 2049)

               # Returns "IP.dst == '192.168.1.50' and TCP.dst_port == 2049"
               # Expression ready for x.match()
               pkt = x.match(expr)
        """
        return "IP.dst == '%s' and TCP.dst_port == %d" % (ipaddr, port)

if __name__ == '__main__':
    # Self test of module
    l_escape = [
        "hello",
        "\x00\\test",
        "single'quote",
        'double"quote',
        'back`quote',
        'single\'double"quote',
        'double"single\'quote',
        'single\'double"back`quote',
        'double"single\'back`quote',
    ]
    ntests = 2*len(l_escape)

    tcount = 0
    for quote in ["'", '"']:
        for data in l_escape:
            expr = "data == %s%s%s" % (quote, Pktt.escape(data), quote)
            expr = re.sub(r'\\\\', r'\\', expr)
            if eval(expr):
                tcount += 1

    if tcount == ntests:
        print "All tests passed!"
        exit(0)
    else:
        print "%d tests failed" % (ntests-tcount)
        exit(1)



#===============================================================================
# Copyright 2015 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Utilities module

Definition for common classes and constants
"""
import os
import nfstest_config as c
from baseobj import BaseObj

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2015 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

# Constants for file type
FTYPE_FILE       = 0 # Regular file
FTYPE_SP_OFFSET  = 1 # Sparse file (write data to offset only)
FTYPE_SP_ZERO    = 2 # Sparse file (write zeros on hole)
FTYPE_SP_DEALLOC = 3 # Sparse file (use deallocate to create holes)

# Space reservation constants
SR_ALLOCATE   = 0 # Allocate
SR_DEALLOCATE = 3 # FALLOC_FL_KEEP_SIZE|FALLOC_FL_PUNCH_HOLE

# Sparse file constants
SP_HOLE = 0 # Hole segment
SP_DATA = 1 # Data segment

SEEK_DATA = 3 # Seek for the next data segment with lseek()
SEEK_HOLE = 4 # Seek for the next hole with lseek()

SEEKmap = {
    SEEK_DATA: "SEEK_DATA",
    SEEK_HOLE: "SEEK_HOLE",
}

def split_path(path):
    """Return list of components in path"""
    ret = os.path.normpath(path).split(os.sep)
    # Remove leading empty component and "." entry
    while len(ret) and ret[0] in ("", "."):
        ret.pop(0)
    return ret

class SparseFile(BaseObj):
    """SparseFile object

       SparseFile() -> New sparse file object

       Usage:
           # Create definition for a sparse file of size 10000 having
           # two holes of size 1000 at offsets 3000 and 6000
           x = SparseFile("/mnt/t/file1", 10000, [3000, 6000], 1000)

           # Object attributes defined after creation using the above
           # sample data:

           # endhole: set to True if the file ends with a hole
           # Above example ends with data so,
           # x.endhole = False

           # data_offsets: list of data segment offsets
           # x.data_offsets = [0, 4000, 7000]

           # hole_offsets: list of hole segment offsets including the
           #   implicit hole at the end of the file
           # x.hole_offsets = [3000, 6000, 10000]

           # sparse_data: list of data/hole segments, each item in the list
           #   has the following format [offset, size, type]
           # x.sparse_data = [[0, 3000, 1], [3000, 1000, 0], [4000, 2000, 1],
           #                  [6000, 1000, 0], [7000, 3000, 1]]
    """
    def __init__(self, absfile, file_size, hole_list, hole_size):
        """Create sparse file object definition, the file is not created
           just the object. Object attributes are defined which makes it
           easy to create the actual file.

           absfile:
               Absolute path name of file
           file_size:
               Total size of sparse file
           hole_list:
               List of hole offsets
           hole_size:
               Size for each hole
        """
        self.filename  = os.path.basename(absfile)
        self.absfile   = absfile
        self.filesize  = file_size
        self.hole_list = hole_list
        self.hole_size = hole_size

        if hole_list[-1] < file_size and hole_list[-1] + hole_size >= file_size:
            # File ends with a hole
            self.endhole = True
        else:
            # File ends with data
            self.endhole = False

        # List of hole offsets
        self.hole_offsets = list(hole_list)

        if not self.endhole:
            # Include the implicit hole at the end of the file
            self.hole_offsets += [file_size]

        # List of data offsets
        self.data_offsets = []
        # List of data and hole segments
        self.sparse_data = []
        if hole_list[0] > 0:
            # There is data at the beginning of the file
            self.data_offsets.append(0)
            self.sparse_data.append([0, hole_list[0], SP_DATA])
        idx = 0
        for offset in hole_list:
            # Append hole segment
            self.sparse_data.append([offset, hole_size, SP_HOLE])
            endhole_offset = offset + hole_size
            if endhole_offset < file_size:
                self.data_offsets.append(endhole_offset)
                if idx == len(hole_list) - 1:
                    # Data segment is up to the end of the file
                    size = file_size - endhole_offset
                else:
                    # Data segment is up to the start of next hole
                    size = hole_list[idx+1] - endhole_offset
                # Append data segment
                self.sparse_data.append([endhole_offset, size, SP_DATA])
            idx += 1



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Test utilities module

Provides a set of tools for testing either the NFS client or the NFS server,
most of the functionality is focused mainly on testing the client.
These tools include the following:

    - Process command line arguments
    - Provide functionality for PASS/FAIL
    - Provide test grouping functionality
    - Provide multiple client support
    - Logging mechanism
    - Debug info control
    - Mount/Unmount control
    - Create files/directories
    - Provide mechanism to start a packet trace
    - Provide mechanism to simulate a network partition
    - Support for pNFS testing

In order to use some of the functionality available, the user id in all the
client hosts must have access to run commands as root using the 'sudo' command
without the need for a password, this includes the host where the test is being
executed. This is used to run commands like 'mount' and 'umount'. Furthermore,
the user id must be able to ssh to remote hosts without the need for a password
if test requires the use of multiple clients.

Network partition is simulated by the use of 'iptables', please be advised
that after every test run the iptables is flushed and reset so any rules
previously setup will be lost. Currently, there is no mechanism to restore
the iptables rules to their original state.
"""
import os
import re
import sys
import time
import fcntl
import ctypes
import struct
import inspect
import textwrap
from utils import *
from formatstr import *
from rexec import Rexec
import nfstest_config as c
from baseobj import BaseObj
from nfs_util import NFSUtil
from optparse import OptionParser, OptionGroup, IndentedHelpFormatter

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.6"

# Constants
PASS = 0
HEAD = 1
INFO = 2
FAIL = -1
WARN = -2
BUG  = -3
IGNR = -4

VT_NORM = "\033[m"
VT_BOLD = "\033[1m"
VT_BLUE = "\033[34m"
VT_HL   = "\033[47m"

_isatty = os.isatty(1)

_test_map = {
    HEAD: "\n*** ",
    INFO: "    ",
    PASS: "    PASS: ",
    FAIL: "    FAIL: ",
    WARN: "    WARN: ",
    BUG:  "    BUG:  ",
    IGNR: "    IGNR: ",
}

# Provide colors on PASS, FAIL, WARN messages
_test_map_c = {
    HEAD: "\n*** ",
    INFO: "    ",
    PASS: "    \033[102mPASS\033[m: ",
    FAIL: "    \033[41m\033[37mFAIL\033[m: ",
    WARN: "    \033[33mWARN\033[m: ",
    BUG:  "    \033[33mBUG\033[m:  ",
    IGNR: "    \033[33mIGNR\033[m: ",
}

_tverbose_map = {'group': 0, 'normal': 1, 'verbose': 2, '0':0, '1':1, '2':2}
_rtverbose_map = dict(zip(_tverbose_map.values(),_tverbose_map))

BaseObj.debug_map(0x100, 'opts', "OPTS: ")

class TestUtil(NFSUtil):
    """TestUtil object

       TestUtil() -> New server object

       Usage:
           x = TestUtil()

           # Process command line options
           x.scan_options()

           # Start packet trace using tcpdump
           x.trace_start()

           # Mount volume
           x.mount()

           # Create file
           x.create_file()

           # Unmount volume
           x.umount()

           # Stop packet trace
           x.trace_stop()

           # Exit script
           x.exit()
    """
    def __init__(self, **kwargs):
        """Constructor

           Initialize object's private data.

           sid:
               Test script ID [default: '']
               This is used to have options targeted for a given ID without
               including these options in any other test script.
           usage:
               Usage string [default: '']
           testnames:
               List of test names [default: []]
               When this list is not empty, the --runtest option is enabled and
               test scripts should use the run_tests() method to run all the
               tests. Test script should have methods named as <testname>_test.
           testgroups:
                Dictionary of test groups where the key is the name of the test
                group and its value is a dictionary having the following keys:
                    tests:
                        A list of tests belonging to this test group
                    desc:
                        Description of the test group, this is displayed
                        in the help if the name of the test group is also
                        included in testnames
                    tincl:
                        Include a comma separated list of tests belonging to
                        this test group to the description [default: False]
                    wrap:
                        Reformat the description so it fits in lines no
                        more than the width given. The description is not
                        formatted for a value of zero [default: 72]

           Example:
               x = TestUtil(testnames=['basic', 'lock'])

               # The following methods should exist:
               x.basic_test()
               x.lock_test()
        """
        self.sid        = kwargs.pop('sid', "")
        self.usage      = kwargs.pop('usage', '')
        self.testnames  = kwargs.pop('testnames', [])
        self.testgroups = kwargs.pop('testgroups', {})
        self.progname = os.path.basename(sys.argv[0])
        self.testname = ""
        if self.progname[-3:] == '.py':
            # Remove extension
            self.progname = self.progname[:-3]
        self._name = None
        self.tverbose = 1
        self._bugmsgs = []
        self.ignore = False
        self.bugmsgs = None
        self.nocleanup = True
        self.isatty = _isatty
        self.test_time = [time.time()]
        self._disp_time = 0
        self._disp_msgs = 0
        self._empty_msg = 0
        self._fileopt = True
        self.remove_list = []
        self.fileidx = 1
        self.diridx = 1
        self.logidx = 1
        self.files = []
        self.dirs = []
        self.test_msgs = []
        self._msg_count = {}
        self._reset_files()
        self._runtest = True
        self.createtraces = False
        self._opts_done = False
        # List of sparse files
        self.sparse_files = []
        # Rexec attributes
        self.rexecobj = None
        self.rexecobj_list = []
        # List of remote files
        self.remote_files = []

        for tid in _test_map:
            self._msg_count[tid] = 0
        self.dindent(4)

        self.optfiles = []
        self.testopts = {}
        NFSUtil.__init__(self)
        self._init_options()

        # Prototypes for libc functions
        self.libc.fallocate.argtypes = ctypes.c_int, ctypes.c_int, ctypes.c_ulong, ctypes.c_ulong
        self.libc.fallocate.restype  = ctypes.c_int

    def __del__(self):
        """Destructor

           Gracefully stop the packet trace, cleanup files, unmount volume,
           and reset network.
        """
        self.debug_repr(0)
        self._tverbose()
        self._print_msg("", single=1)
        count = self.dprint_count()
        self.dprint('DBG7', "Calling %s() destructor" % self.__class__.__name__)
        self.trace_stop()
        self.cleanup(newline=False)
        # Call base destructor
        NFSUtil.__del__(self)
        if self.dprint_count() > count:
            self._empty_msg = 0

        if len(self.test_msgs) > 0:
            if getattr(self, 'logfile', None):
                if not self._empty_msg:
                    print ""
                print "Logfile: %s" % self.logfile
                self._empty_msg = 0
            ntests, tmsg = self._total_counts(self._msg_count)
            if ntests > 0:
                self._print_msg("", single=1)
                msg = "%d tests%s" % (ntests, tmsg)
                self.write_log(msg)
                if self._msg_count[FAIL] > 0:
                    msg = "\033[31m" + msg + "\033[m" if self.isatty else msg
                elif self._msg_count[WARN] > 0:
                    msg = "\033[33m" + msg + "\033[m" if self.isatty else msg
                else:
                    msg = "\033[32m" + msg + "\033[m" if self.isatty else msg
                print msg
        if self._opts_done:
            self.total_time = time.time() - self.test_time[0]
            total_str = "\nTotal time: %s" % self._print_time(self.total_time)
            self.write_log(total_str)
            print total_str
        self.close_log()

    def _verify_testnames(self):
        """Process --runtest option."""
        if not hasattr(self, 'runtest'):
            return
        if self.runtest == 'all':
            self.testlist = self.testnames
        else:
            if self.runtest[0] == '^':
                # List is negated tests -- do not run the tests listed
                runtest = self.runtest.replace('^', '', 1)
                negtestlist = self.str_list(runtest)
                self.testlist = list(self.testnames)
                for testname in negtestlist:
                    if testname in self.testlist:
                        self.testlist.remove(testname)
                    elif testname in self.testgroups:
                        # Remove all tests in the test group
                        for tname in self.testgroups[testname].get("tests", []):
                            if tname in self.testlist:
                                self.testlist.remove(tname)
                    elif testname not in self.testnames:
                        self.opts.error("invalid value given --runtest=%s" % self.runtest)
            else:
                idx = 0
                self.testlist = self.str_list(self.runtest)
                # Process the test groups by including all the tests
                # in the test group to the list of tests to run
                for testname in self.testlist:
                    tgroup = self.testgroups.get(testname)
                    if tgroup is not None:
                        # Add tests from the test group to the list
                        self.testlist.remove(testname)
                        for tname in tgroup.get("tests", []):
                            self.testlist.insert(idx, tname)
                            idx += 1
                    idx += 1
            if self.testlist is None:
                self.opts.error("invalid value given --runtest=%s" % self.runtest)
        msg = ''
        for testname in self.testlist:
            if testname not in self.testnames:
                msg += "Invalid test name:    %s\n" % testname
            elif not hasattr(self, testname + '_test'):
                msg += "Test not implemented: %s\n" % testname
            else:
                tname = testname + '_test'
        if len(msg) > 0:
            self.config(msg)

    def _init_options(self):
        """Initialize command line options parsing and definitions."""
        self.opts = OptionParser("%prog [options]", formatter = IndentedHelpFormatter(2, 10), version = "%prog " + __version__)
        hmsg = "File where options are specified besides the system wide " + \
               "file /etc/nfstest, user wide file $HOME/.nfstest or in " + \
               "the current directory .nfstest file"
        self.opts.add_option("-f", "--file", default="", help=hmsg)

        self.nfs_opgroup = OptionGroup(self.opts, "NFS specific options")
        hmsg = "Server name or IP address"
        self.nfs_opgroup.add_option("-s", "--server", default=self.server, help=hmsg)
        hmsg = "Exported file system to mount [default: '%default']"
        self.nfs_opgroup.add_option("-e", "--export", default=self.export, help=hmsg)
        hmsg = "NFS version, e.g., 3, 4, 4.1, etc. [default: %default]"
        self.nfs_opgroup.add_option("--nfsversion", type="float", default=self.nfsversion, help=hmsg)
        hmsg = "Mount point [default: '%default']"
        self.nfs_opgroup.add_option("-m", "--mtpoint", default=self.mtpoint, help=hmsg)
        hmsg = "NFS server port [default: %default]"
        self.nfs_opgroup.add_option("-p", "--port", type="int", default=self.port, help=hmsg)
        hmsg = "NFS protocol name [default: '%default']"
        self.nfs_opgroup.add_option("--proto", default=self.proto, help=hmsg)
        hmsg = "Security flavor [default: '%default']"
        self.nfs_opgroup.add_option("--sec", default=self.sec, help=hmsg)
        hmsg = "Mount options [default: '%default']"
        self.nfs_opgroup.add_option("-o", "--mtopts", default=self.mtopts, help=hmsg)
        hmsg = "Data directory where files are created, directory is " + \
               "created on the mount point [default: '%default']"
        self.nfs_opgroup.add_option("--datadir", default=self.datadir, help=hmsg)
        self.opts.add_option_group(self.nfs_opgroup)

        self.log_opgroup = OptionGroup(self.opts, "Logging options")
        hmsg = "Verbose level for debug messages [default: '%default']"
        self.log_opgroup.add_option("-v", "--verbose", default="none", help=hmsg)
        hmsg = "Verbose level for test messages [default: '%default']"
        self.log_opgroup.add_option("--tverbose", default=_rtverbose_map[self.tverbose], help=hmsg)
        hmsg = "Create log file"
        self.log_opgroup.add_option("--createlog", action="store_true", default=False, help=hmsg)
        hmsg = "Create rexec log files"
        self.log_opgroup.add_option("--rexeclog", action="store_true", default=False, help=hmsg)
        hmsg = "Display warnings"
        self.log_opgroup.add_option("--warnings", action="store_true", default=False, help=hmsg)
        hmsg = "Informational tag, it is displayed as an INFO message [default: '%default']"
        self.log_opgroup.add_option("--tag", default="", help=hmsg)
        hmsg = "Do not use terminal colors on output"
        self.log_opgroup.add_option("--notty", action="store_true", default=False, help=hmsg)
        self.opts.add_option_group(self.log_opgroup)

        self.cap_opgroup = OptionGroup(self.opts, "Packet trace options")
        hmsg = "Create a packet trace for each test"
        self.cap_opgroup.add_option("--createtraces", action="store_true", default=False, help=hmsg)
        hmsg = "Capture buffer size for tcpdump [default: %default]"
        self.cap_opgroup.add_option("--tbsize", default="192k", help=hmsg)
        hmsg = "Seconds to delay before stopping packet trace [default: %default]"
        self.cap_opgroup.add_option("--trcdelay", type="float", default=0.0, help=hmsg)
        hmsg = "Do not remove any trace files [default: remove trace files if no errors]"
        self.cap_opgroup.add_option("--keeptraces", action="store_true", default=False, help=hmsg)
        hmsg = "Remove trace files [default: remove trace files if no errors]"
        self.cap_opgroup.add_option("--rmtraces", action="store_true", default=False, help=hmsg)
        hmsg = "Device interface [default: automatically selected]"
        self.cap_opgroup.add_option("-i", "--interface", default=None, help=hmsg)
        self.opts.add_option_group(self.cap_opgroup)

        self.file_opgroup = OptionGroup(self.opts, "File options")
        hmsg = "Number of files to create [default: %default]"
        self.file_opgroup.add_option("--nfiles", type="int", default=2, help=hmsg)
        hmsg = "File size to use for test files [default: %default]"
        self.file_opgroup.add_option("--filesize", default="64k", help=hmsg)
        hmsg = "Read size to use when reading files [default: %default]"
        self.file_opgroup.add_option("--rsize", default="4k", help=hmsg)
        hmsg = "Write size to use when writing files [default: %default]"
        self.file_opgroup.add_option("--wsize", default="4k", help=hmsg)
        hmsg = "Seconds to delay I/O operations [default: %default]"
        self.file_opgroup.add_option("--iodelay", type="float", default=0.1, help=hmsg)
        hmsg = "Read/Write offset delta [default: %default]"
        self.file_opgroup.add_option("--offset-delta", default="4k", help=hmsg)
        self.opts.add_option_group(self.file_opgroup)

        self.path_opgroup = OptionGroup(self.opts, "Path options")
        hmsg = "Full path of binary for sudo [default: '%default']"
        self.path_opgroup.add_option("--sudo", default=self.sudo, help=hmsg)
        hmsg = "Full path of binary for tcpdump [default: '%default']"
        self.path_opgroup.add_option("--tcpdump", default=self.tcpdump, help=hmsg)
        hmsg = "Full path of binary for iptables [default: '%default']"
        self.path_opgroup.add_option("--iptables", default=self.iptables, help=hmsg)
        hmsg = "Full path of log messages file [default: '%default']"
        self.path_opgroup.add_option("--messages", default=self.messages, help=hmsg)
        hmsg = "Temporary directory [default: '%default']"
        self.path_opgroup.add_option("--tmpdir", default=self.tmpdir, help=hmsg)
        self.opts.add_option_group(self.path_opgroup)

        self.dbg_opgroup = OptionGroup(self.opts, "Debug options")
        hmsg = "Do not cleanup created files"
        self.dbg_opgroup.add_option("--nocleanup", action="store_true", default=False, help=hmsg)
        hmsg = "File containing test messages to mark as bugs if they failed"
        self.dbg_opgroup.add_option("--bugmsgs", default=self.bugmsgs, help=hmsg)
        hmsg = "Ignore all bugs given by bugmsgs"
        self.dbg_opgroup.add_option("--ignore", action="store_true", default=self.ignore, help=hmsg)
        hmsg = "Do not mount server and run the tests on local disk space"
        self.dbg_opgroup.add_option("--nomount", action="store_true", default=self.nomount, help=hmsg)
        hmsg = "Base name for all files and logs [default: automatically generated]"
        self.dbg_opgroup.add_option("--basename", default='', help=hmsg)
        hmsg = "Set NFS kernel debug flags and save log messages [default: '%default']"
        self.dbg_opgroup.add_option("--nfsdebug", default=self.nfsdebug, help=hmsg)
        hmsg = "Set RPC kernel debug flags and save log messages [default: '%default']"
        self.dbg_opgroup.add_option("--rpcdebug", default=self.rpcdebug, help=hmsg)
        hmsg = "Display main packets related to the given test"
        self.dbg_opgroup.add_option("--pktdisp", action="store_true", default=False, help=hmsg)
        self.opts.add_option_group(self.dbg_opgroup)

        usage = self.usage
        if len(self.testnames) > 0:
            self.test_opgroup = OptionGroup(self.opts, "Test options")
            hmsg = "Comma separated list of tests to run, if list starts " + \
                   "with a '^' then all tests are run except the ones " + \
                   "listed [default: '%default']"
            self.test_opgroup.add_option("--runtest", default='all', help=hmsg)
            self.opts.add_option_group(self.test_opgroup)
            if len(usage) == 0:
                usage = "%prog [options]"
            usage += "\n\nAvailable tests:"
            for tgname, item in self.testgroups.items():
                tlist = item.get("tests", [])
                tincl = item.get("tincl", True)
                wrap = item.get("wrap", 72)
                if item.get("desc", None) is not None:
                    if tincl and tlist:
                        # Add the list of tests for this test group
                        # to the description
                        item["desc"] += ", ".join(tlist)
                    if wrap > 0:
                        item["desc"] = "\n".join(textwrap.wrap(item["desc"], wrap))
            for tname in self.testnames:
                tgroup = self.testgroups.get(tname)
                desc = None
                if tgroup is not None:
                    desc = tgroup.get("desc")
                if desc is None:
                    desc = getattr(self, tname+'_test').__doc__
                if desc is not None:
                    lines = desc.lstrip().split('\n')
                    desc = lines.pop(0)
                    if len(desc) > 0:
                        desc += '\n'
                    desc += textwrap.dedent("\n".join(lines))
                    desc = desc.replace("\n", "\n        ").rstrip()
                usage += "\n    %s:\n        %s\n" % (tname, desc)
            usage = usage.rstrip()
            # Remove test group names from the list of tests
            for tname in self.testgroups:
                self.testnames.remove(tname)
        if len(usage) > 0:
            self.opts.set_usage(usage)
        self._cmd_line = " ".join(sys.argv)
        self._opts = {}

    @staticmethod
    def str_list(value, type=str, sep=","):
        """Return a list of <type> elements from the comma separated string."""
        slist = []
        try:
            for item in value.replace(' ', '').split(sep):
                if len(item) > 0:
                    if type == int:
                        val = int(item)
                    elif type == float:
                        val = float(item)
                    else:
                        val = item
                    slist.append(val)
                else:
                    slist.append(None)
        except:
            return
        return slist

    @staticmethod
    def get_list(value, hash, type=str):
        """Return a list of elements from the comma separated string.
           Validate and translate these elements using the input dictionary
           'hash' where every element in the string is the key of 'hash'
           and its value is appended to the returned list.
        """
        rlist = []
        slist = TestUtil.str_list(value)
        if slist is None:
            return
        for i_item in slist:
            item = i_item.lower()
            if hash.has_key(item):
                rlist.append(hash[item])
            else:
                return
        return rlist

    def scan_options(self):
        """Process command line options.

           Process all the options in the file given by '--file', then the
           ones in the command line. This allows for command line options
           to over write options given in the file.

           Format of options file:
               # For options expecting a value
               <option_name> = <value>

               # For boolean (flag) options
               <option_name>

           Process options files and make sure not to process the same file
           twice, this is used for the case where HOMECFG and CWDCFG are the
           same, more specifically when environment variable HOME is not
           defined. Also, the precedence order is defined as follows:
             1. options given in command line
             2. options given in file specified by the -f|--file option
             3. options given in file specified by ./.nfstest
             4. options given in file specified by $HOME/.nfstest
             5. options given in file specified by /etc/nfstest

           NOTE:
             Must use the long name of the option (--<option_name>) in the file.
        """
        opts, args = self.opts.parse_args()
        if self._fileopt:
            # Find which options files exist and make sure not to process the
            # same file twice, this is used for the case where HOMECFG and
            # CWDCFG are the same, more specifically when environment variable
            # HOME is not defined.
            ofiles = {}
            self.optfiles = [[opts.file, []]] if opts.file else []
            for optfile in [c.NFSTEST_CWDCFG, c.NFSTEST_HOMECFG, c.NFSTEST_CONFIG]:
                if ofiles.get(optfile) is None:
                    # Add file if it has not been added yet
                    ofiles[optfile] = 1
                    if os.path.exists(optfile):
                        self.optfiles.insert(0, [optfile, []])
        if self.optfiles and self._fileopt:
            # Options are given in any of the options files
            self._fileopt = False # Only process the '--file' option once
            argv = []
            for (optfile, lines) in self.optfiles:
                bcount = 0
                islist = False
                idblock = None
                testblock = None
                for optline in open(optfile, 'r'):
                    line = optline.strip()
                    if len(line) == 0 or line[0] == '#':
                        # Skip comments
                        continue
                    # Save current line of file for displaying purposes
                    lines.append(optline.rstrip())
                    # Process valid options, option name and value is separated
                    # by spaces or an equal sign
                    m = re.search("([^=\s]+)\s*=?\s*(.*)", line)
                    name = m.group(1)
                    name = name.strip()
                    value = m.group(2)
                    # Add current option to argument list as if the option was
                    # given on the command line to be able to use parse_args()
                    # again to process all options given in the options files
                    if name in ["}", "]"]:
                        # End of block, make sure to close an opened testblock
                        # first before closing an opened idblock
                        bcount -= 1
                        if testblock is not None:
                            testblock = None
                        else:
                            idblock = None
                    elif len(value) > 0:
                        value = value.strip()
                        if value in ["{", "["]:
                            # Start of block, make sure to open an idblock
                            # first before opening a testblock
                            islist = True if value == "[" else False
                            bcount += 1
                            if idblock is None:
                                idblock = name
                            elif idblock == self.sid:
                                # Open a testblock only if testblock is located
                                # inside an idblock corresponding to script ID
                                testblock = name
                                if self.testopts.get(name) is None:
                                    # Initialize testblock only if it has not
                                    # been initialized, this allows for multiple
                                    # definitions of the same testblock
                                    if islist:
                                        self.testopts[name] = []
                                    else:
                                        self.testopts[name] = {}
                        elif testblock is not None:
                            # Inside a testblock, add name/value to testblock
                            # dictionary
                            if islist:
                                self.testopts[testblock].append(line)
                            else:
                                self.testopts[testblock][name] = value
                        elif idblock is None or idblock == self.sid:
                            # Include all general options and options given
                            # by the block specified by the correct script ID
                            argv.append("--%s=%s" % (name, value))
                    elif testblock is not None:
                        # Inside a testblock, add name to testblock dictionary
                        if islist:
                            self.testopts[testblock].append(name)
                        else:
                            self.testopts[testblock][name] = True
                    elif idblock is None or (idblock == self.sid and testblock is None):
                        # Include all general options and options given
                        # by the block specified by the correct script ID
                        argv.append("--%s" % name)
                if bcount != 0:
                    self.config("Missing closing brace in options file '%s'" % optfile)
            # Add all other options in the command line, make sure all options
            # explicitly given in the command line have higher precedence than
            # options given in any of the options files
            sys.argv[1:] = argv + sys.argv[1:]
            # Process the command line arguments again to overwrite options
            # explicitly given in the command line in conjunction with the
            # --file option
            self.scan_options()
        else:
            if opts.notty:
                # Do not use terminal colors
                self.isatty = False

            try:
                # Set verbose level mask
                self.debug_level(opts.verbose)
            except Exception, e:
                self.opts.error("Invalid verbose level <%s>: %s" % (opts.verbose, e))

            if opts.createlog and len(opts.basename) == 0:
                self.logfile = "%s/%s.log" % (opts.tmpdir, self.get_name())
                self.open_log(self.logfile)

            _lines = [self._cmd_line]
            for (optfile, lines) in self.optfiles:
                # Add the content of each option file that has been processed
                if len(lines) > 0:
                    _lines.append("")
                    _lines.append("Contents of options file [%s]:" % optfile)
                    _lines += lines
            self.dprint('OPTS', "\n".join(_lines))
            self.dprint('OPTS', "")
            for key in sorted(vars(opts)):
                value = getattr(opts,key)
                self._opts[key] = value
                line = "%s = %s" % (key, value)
                self.dprint('OPTS', line)
            self.dprint('OPTS', "")

            if len(opts.tag) > 0:
                # Display tag information
                self.dprint('INFO', "TAG: %s" % opts.tag)

            # Display system information
            self.dprint('INFO', "SYSTEM: %s" % " ".join(os.uname()))

            # Process all command line arguments -- all will be part of the
            # objects namespace
            self.__dict__.update(opts.__dict__)
            if not self.server:
                self.opts.error("server option is required")
            self._verify_testnames()
            ipv6 = self.proto[-1] == '6'
            # Get IP address of server
            self.server_ipaddr = self.get_ip_address(host=self.server, ipv6=ipv6)
            # Get IP address of client
            self.client_ipaddr = self.get_ip_address(ipv6=ipv6)
            if self.interface is None:
                out = self.get_route(self.server_ipaddr)
                if out[1] is not None:
                    self.interface = out[1]
                    if out[2] is not None:
                        self.client_ipaddr = out[2]
                else:
                    self.interface = c.NFSTEST_INTERFACE
            self.ipaddr = self.client_ipaddr

            self.tverbose = _tverbose_map.get(self.tverbose)
            if self.tverbose is None:
                self.opts.error("invalid value for tverbose option")

            # Convert units
            self.filesize     = int_units(self.filesize)
            self.rsize        = int_units(self.rsize)
            self.wsize        = int_units(self.wsize)
            self.offset_delta = int_units(self.offset_delta)
            self.tbsize       = int_units(self.tbsize)

            if len(self.basename) > 0:
                self._name      = self.basename
                self.nomount    = True
                self.notrace    = True
                self.keeptraces = True
            if self.bugmsgs is not None:
                try:
                    for line in open(self.bugmsgs, 'r'):
                        line = line.strip()
                        if len(line):
                            self._bugmsgs.append(line)
                except Exception as e:
                    self.config("Unable to load bug messages from file '%s': %r" % (self.bugmsgs, e))

            # Set base name for trace files and log message files
            self.tracename = self.get_name()
            self.dbgname = self.get_name()

            # Make sure the network is reset
            self.network_reset()
            self._opts_done = True

    def test_options(self, name=None):
        """Get options for the given test name. If the test name is not given
           it is determined by inspecting the stack to find which method is
           calling this method.
        """
        if name is None:
            # Get current testname
            name = self.testname
            if len(name) == 0:
                # Get correct test name by inspecting the stack to find which
                # method is calling this method
                out = inspect.stack()
                name = out[1][3].replace("_test", "")

        # Get options given for this specific test name
        opts = self.testopts.get(name, {})

        # Find if any of the test options are regular expressions
        for key in self.testopts.keys():
            m = re.search("^re\((.*)\)$", key)
            if m:
                # Regular expression specified by re()
                regex = m.group(1)
            else:
                # Find if regular expression is specified by the characters
                # used in the name
                m = re.search("[.^$?+\\\[\]()|]", key)
                regex = key
            if m and re.search(regex, name):
                # Key is specified as a regular expression and matches
                # the test name given, add these options to any options
                # already given by static name match making sure the
                # options given by the exact name are not overwritten
                # by the ones found from a regular expression
                opts = dict(self.testopts[key].items() + opts.items())
        return opts

    def get_logname(self):
        """Get next log file name."""
        logfile = "%s/%s_%d.log" % (self.tmpdir, self.get_name(), self.logidx)
        self.logidx += 1
        return logfile

    def setup(self, nfiles=None):
        """Set up test environment.

           Create nfiles number of files [default: --nfiles option]
        """
        self.dprint('DBG7', "SETUP starts")
        if nfiles is None:
            nfiles = self.nfiles
        need_umount = False
        if not self.mounted and nfiles > 0:
            need_umount = True
            self.umount()
            self.mount()

        # Create files
        for i in range(nfiles):
            self.create_file()

        if need_umount:
            self.umount()
        self.dprint('DBG7', "SETUP done")

    def cleanup(self, newline=True):
        """Clean up test environment.

           Remove any files created: test files, trace files.
        """
        if self.nocleanup:
            # Nothing to clean up
            return
        # Cleanup just once
        self.nocleanup = True

        if newline:
            self._tverbose()
            self._print_msg("", single=1)
        self.dprint('DBG7', "CLEANUP starts")
        if not self.mounted and self.remove_list:
            self.mount()

        for rexecobj in self.rexecobj_list:
            try:
                if rexecobj.remote:
                    srvname = "at %s" % rexecobj.servername
                else:
                    srvname = "locally"
                self.dprint('DBG3', "    Stop remote procedure server %s" % srvname)
                rexecobj.close()
            except:
                pass
        self.rexecobj = None
        self.rexecobj_list = []

        for item in self.remote_files:
            try:
                cmd = "scp %s:%s %s" % (item[0], item[1], self.tmpdir)
                self.run_cmd(cmd, dlevel='DBG4', msg="    Copy remote file: ")
            except Exception as e:
                self.dprint('DBG7', "    ERROR: %s" % e)

        for item in self.remote_files:
            try:
                cmd = "ssh -t %s sudo rm -f %s" % (item[0], item[1])
                self.run_cmd(cmd, dlevel='DBG4', msg="    Removing remote file: ")
            except:
                pass

        if not self.keeptraces and (self.rmtraces or self._msg_count[FAIL] == 0):
            for rfile in self.tracefiles:
                try:
                    # Remove trace files as root
                    self.dprint('DBG5', "    Removing trace file [%s]" % rfile)
                    os.system(self.sudo_cmd("rm -f %s" % rfile))
                except:
                    pass

        for rfile in reversed(self.remove_list):
            try:
                if os.path.exists(rfile):
                    if os.path.isfile(rfile):
                        self.dprint('DBG5', "    Removing file [%s]" % rfile)
                        os.unlink(rfile)
                    elif os.path.islink(rfile):
                        self.dprint('DBG5', "    Removing symbolic link [%s]" % rfile)
                        os.unlink(rfile)
                    elif os.path.isdir(rfile):
                        self.dprint('DBG5', "    Removing directory [%s]" % rfile)
                        os.rmdir(rfile)
                    else:
                        self.dprint('DBG5', "    Removing [%s]" % rfile)
                        os.unlink(rfile)
            except:
                pass
        self.umount()
        self.dprint('DBG7', "CLEANUP done")

    def create_rexec(self, servername=None, **kwds):
        """Create remote server object."""
        if self.rexeclog:
            kwds["logfile"] = kwds.get("logfile", self.get_logname())
        else:
            kwds["logfile"] = None

        # Start remote procedure server on given client
        if servername in [None, "", "localhost", "127.0.0.1"]:
            svrname = "locally"
        else:
            svrname = "at %s" % servername
            if kwds.get("logfile") is not None:
                self.remote_files.append([servername, kwds["logfile"]])

        self.dprint('DBG2', "Start remote procedure server %s" % svrname)
        self.rexecobj = Rexec(servername, **kwds)
        self.rexecobj_list.append(self.rexecobj)
        return self.rexecobj

    def run_tests(self, **kwargs):
        """Run all test specified by the --runtest option.

           testnames:
               List of testnames to run [default: all tests given by --testnames]

           All other arguments given are passed to the test methods.
        """
        testnames = kwargs.pop("testnames", self.testlist)
        for name in self.testlist:
            testmethod = name + '_test'
            if name in testnames and hasattr(self, testmethod):
                self._runtest = True
                self._tverbose()
                # Set current testname on object
                self.testname = name
                # Execute test
                getattr(self, testmethod)(**kwargs)

    def _print_msg(self, msg, tid=None, single=0):
        """Display message to the screen and to the log file."""
        if single and self._empty_msg:
            # Display only a single empty line
            return
        tidmsg_l = '' if tid is None else _test_map[tid]
        self.write_log(tidmsg_l + msg)
        if self.isatty:
            tidmsg_s = _test_map_c.get(tid, tidmsg_l)
            if tid == HEAD:
                msg = VT_HL + VT_BOLD + msg + VT_NORM
            elif tid == INFO:
                msg = VT_BLUE + VT_BOLD + msg + VT_NORM
            elif tid in [PASS, FAIL]:
                msg = VT_BOLD + msg + VT_NORM
        else:
            tidmsg_s = tidmsg_l
        print tidmsg_s + msg
        sys.stdout.flush()
        if len(msg) > 0:
            self._empty_msg = 0
            self._disp_msgs += 1
        else:
            self._empty_msg = 1

    def _print_time(self, sec):
        """Return the given time in the format [[%dh]%dm]%fs."""
        hh = int(sec)/3600
        sec -= 3600.0*hh
        mm = int(sec)/60
        sec -= 60.0*mm
        ret = "%fs" % sec
        if mm > 0:
            ret = "%dm%s" % (mm, ret)
        if hh > 0:
            ret = "%dh%s" % (hh, ret)
        return ret

    def _total_counts(self, gcounts):
        """Internal method to return a string containing how many tests passed
           and how many failed.
        """
        total = gcounts[PASS] + gcounts[FAIL] + gcounts[BUG]
        bugs  = ", %d known bugs" % gcounts[BUG]  if gcounts[BUG] > 0  else ""
        warns = ", %d warnings"   % gcounts[WARN] if gcounts[WARN] > 0 else ""
        tmsg = " (%d passed, %d failed%s%s)" % (gcounts[PASS], gcounts[FAIL], bugs, warns)
        return (total, tmsg)

    def _tverbose(self):
        """Display test group message as a PASS/FAIL including the number
           of tests that passed and failed within this test group when the
           tverbose option is set to 'group' or level 0. It also groups all
           test messages belonging to the same sub-group when the tverbose
           option is set to 'normal' or level 1.
        """
        if self.tverbose == 0 and len(self.test_msgs) > 0:
            # Get the count for each type of message within the
            # current test group
            gcounts = {}
            for tid in _test_map:
                gcounts[tid] = 0
            for item in self.test_msgs[-1]:
                if item[3]:
                    # This message has already been displayed
                    continue
                item[3] = 1
                if len(item[2]) > 0:
                    # Include all subtest results on the counts
                    for subitem in item[2]:
                        gcounts[subitem[0]] += 1
                else:
                    # No subtests, include just the test results
                    gcounts[item[0]] += 1
            (total, tmsg) = self._total_counts(gcounts)
            if total > 1:
                # Fail the current test group if at least one of the tests within
                # this group fails
                tid = FAIL if gcounts[FAIL] > 0 else PASS
                # Just add the test group as a single test entity in the total count
                self._msg_count[tid] += 1
                # Just display the test group message with the count of tests
                # that passed and failed within this test group
                msg = self.test_msgs[-1][0][1].replace("\n", "\n          ")
                self._print_msg(msg + tmsg, tid)
                sys.stdout.flush()
        elif self.tverbose == 1 and len(self.test_msgs) > 0:
            # Process all sub-groups within the current test group
            group = self.test_msgs[-1]
            for subgroup in group:
                sgtid = subgroup[0]
                msg = subgroup[1]
                subtests = subgroup[2]
                disp = subgroup[3]
                if len(subtests) == 0 or disp:
                    # Nothing to process, there are no subtests
                    # or have already been displayed
                    continue
                # Do not display message again
                subgroup[3] = 1
                # Get the count for each type of message within this
                # test sub-group
                gcounts = {}
                for tid in _test_map:
                    gcounts[tid] = 0
                for subtest in subtests:
                    gcounts[subtest[0]] += 1
                (total, tmsg) = self._total_counts(gcounts)
                # Just add the test sub-group as a single test entity in the
                # total count
                self._msg_count[sgtid] += 1
                # Just display the test group message with the count of tests
                # that passed and failed within this test group
                msg = msg.replace("\n", "\n          ")
                self._print_msg(msg + tmsg, sgtid)
                sys.stdout.flush()
        if self.createtraces:
            self.trace_stop()
        self._test_time()

    def _subgroup_id(self, subgroup, tid, subtest):
        """Internal method to return the index of the sub-group message"""
        index = 0
        grpid = None
        # Search the given message in all the sub-group messages
        # within the current group
        group = self.test_msgs[-1]
        if subtest is not None:
            # Look for sub-group message only if this test has subtests
            for item in group:
                if subgroup == item[1]:
                    # Sub-group message found
                    grpid = index
                    break
                index += 1
        if grpid is None:
            # Sub-group not found, add it
            # [tid, test-message, list-of-subtest-results]
            grpid = len(group)
            group.append([tid, subgroup, [], 0])
        return grpid

    def _test_msg(self, tid, msg, subtest=None, failmsg=None):
        """Common method to display and group test messages."""
        if len(self.test_msgs) == 0 or tid == HEAD:
            # This is the first test message or the start of a group,
            # so process the previous group if any and create a placeholder
            # for the current group
            if not self._runtest:
                self._tverbose()
            self.test_msgs.append([])
        # Match the given message to a sub-group or add it if no match
        grpid = self._subgroup_id(msg, tid, subtest)
        if subtest is not None:
            # A subtest is given so added to the proper sub-group
            subgroup = self.test_msgs[-1][grpid]
            subgroup[2].append([tid, subtest])
            if subgroup[0] == PASS and tid == FAIL:
                # Subtest failed so fail the subgroup
                subgroup[0] = FAIL
        if self.tverbose == 2 or (self.tverbose == 1 and subtest is None):
            # Display the test message if tverbose flag is set to verbose(2)
            # or if there is no subtest when tverbose is set to normal(1)
            self._msg_count[tid] += 1
            if subtest is not None:
                msg += subtest
            if failmsg is not None and tid == FAIL:
                msg += failmsg
            msg = msg.replace("\n", "\n          ")
            self._print_msg(msg, tid)

        if tid == HEAD:
            if self._runtest:
                self.test_info("TEST: Running test '%s'" % self.testname)
            self._runtest = False
            if self.createtraces:
                self.trace_start()

    def _test_time(self):
        """Add an INFO message having the time difference between the current
           time and the time of the last call to this method.
        """
        if self._disp_time >= self._disp_msgs + self.dprint_count():
            return
        self.test_time.append(time.time())
        if self._opts_done and len(self.test_time) > 1:
            ttime = self.test_time[-1] - self.test_time[-2]
            self._test_msg(INFO, "TIME: %s" % self._print_time(ttime))
        self._disp_time = self._disp_msgs + self.dprint_count()

    def exit(self):
        """Terminate script with an exit value of 0 when all tests passed
           and a value of 1 when there is at least one test failure.
        """
        if self._msg_count[FAIL] > 0:
            sys.exit(1)
        else:
            sys.exit(0)

    def config(self, msg):
        """Display config message and terminate test with an exit value of 2."""
        msg = "CONFIG: " + msg
        msg = msg.replace("\n", "\n        ")
        self.write_log(msg)
        print msg
        sys.exit(2)

    def test_info(self, msg):
        """Display info message."""
        self._test_msg(INFO, msg)

    def test_group(self, msg):
        """Display heading message and start a test group.

           If tverbose=group or level 0:
               Group message is displayed as a PASS/FAIL message including the
               number of tests that passed and failed within this test group.
           If tverbose=normal|verbose or level 1|2:
               Group message is displayed as a heading messages for the tests
               belonging to this test group.
        """
        self._test_msg(HEAD, msg)

    def warning(self, msg):
        """Display warning message."""
        if self.warnings:
            self._test_msg(WARN, msg)

    def test(self, expr, msg, subtest=None, failmsg=None, terminate=False):
        """Test expr and display message as PASS/FAIL, terminate execution
           if terminate option is True.

           expr:
               If expr is true, display as a PASS message,
               otherwise as a FAIL message
           msg:
               Message to display
           subtest:
               If given, append this string to the displayed message and
               mark this test as a member of the sub-group given by msg
           failmsg:
               If given, append this string to the displayed message when
               expr is false [default: None]
           terminate:
               Terminate execution if true and expr is false [default: False]

           If tverbose=normal or level 1:
               Sub-group message is displayed as a PASS/FAIL message including
               the number of tests that passed and failed within the sub-group
           If tverbose=verbose or level 2:
               All tests messages are displayed
        """
        tid = PASS if expr else FAIL
        if len(self._bugmsgs):
            for tmsg in self._bugmsgs:
                if re.search(tmsg, msg):
                    if self.ignore:
                        # Do not count as a failure if bugmsgs and ignore are set
                        # and it is a failure
                        tid = IGNR if tid == FAIL else tid
                    else:
                        # Do not count as a failure if bugmsgs is set and it is a failure
                        tid = BUG  if tid == FAIL else tid
                        # Count as a failure if bugmsgs is set and the test actually passed
                        tid = FAIL if tid == PASS else tid
                        if tid == FAIL:
                            failmsg = " -- test actually PASSed but failing because --bugmsgs is used"
                    break
        self._test_msg(tid, msg, subtest=subtest, failmsg=failmsg)
        if tid == FAIL and terminate:
            self.exit()

    def testid_count(self, tid):
        """Return the number of instances the testid has occurred."""
        return self._msg_count[tid]

    def get_name(self):
        """Get unique name for this instance."""
        if not self._name:
            timestr = self.timestamp("{0:date:%Y%m%d%H%M%S}")
            self._name = "%s_%s" % (self.progname, timestr)
        return self._name

    def get_dirname(self, dir=None):
        """Return a unique directory name under the given directory."""
        self.dirname = "%s_d_%d" % (self.get_name(), self.diridx)
        self.diridx += 1
        self.absdir = self.abspath(self.dirname, dir=dir)
        self.dirs.append(self.dirname)
        self.remove_list.append(self.absdir)
        return self.dirname

    def get_filename(self, dir=None):
        """Return a unique file name under the given directory."""
        self.filename = "%s_f_%d" % (self.get_name(), self.fileidx)
        self.fileidx += 1
        self.absfile = self.abspath(self.filename, dir=dir)
        self.files.append(self.filename)
        self.remove_list.append(self.absfile)
        return self.filename

    def data_pattern(self, offset, size, pattern=None):
        """Return data pattern.

           offset:
               Starting offset of pattern
           size:
               Size of data to return
           pattern:
               Data pattern to return, default is of the form:
               hex_offset(0x%08X) abcdefghijklmnopqrst\\n
        """
        data = ''
        if pattern is None:
            pattern = 'abcdefghijklmnopqrst'
            line_len = 32
            default = True
        else:
            line_len = len(pattern)
            default = False

        s_offset = offset % line_len
        offset = offset - s_offset
        N = int(0.9999 + (size + s_offset) / float(line_len))

        for i in range(0,N):
            if default:
                str_offset = "0x%08X " % offset
                plen = 31 - len(str_offset)
                data += str_offset + pattern[:plen] + '\n'
                offset += line_len
            else:
                data += pattern
        return data[s_offset:size+s_offset]

    def delay_io(self, delay=None):
        """Delay I/O by value given or the value given in --iodelay option."""
        if delay is None:
            delay = self.iodelay
        if not self.nomount and len(self.basename) == 0:
            # Slow down traffic for tcpdump to capture all packets
            time.sleep(delay)

    def create_dir(self, dir=None, mode=0755):
        """Create a directory under the given directory with the given mode."""
        self.get_dirname(dir=dir)
        self.dprint('DBG3', "Creating directory [%s]" % self.absdir)
        os.mkdir(self.absdir, mode)
        return self.dirname

    def write_data(self, fd, offset=0, size=None, pattern=None):
        """Write data to the file given by the file descriptor

           fd:
               File descriptor
           offset:
               File offset where data will be written to [default: 0]
           size:
               Total number of bytes to write [default: --filesize option]
           pattern:
               Data pattern to write to the file [default: data_pattern default]
        """
        if size is None:
            size = self.filesize

        while size > 0:
            # Write as much as wsize bytes per write call
            dsize = min(self.wsize, size)
            os.lseek(fd, offset, 0)
            count = os.write(fd, self.data_pattern(offset, dsize, pattern))
            size -= count
            offset += count

    def create_file(self, offset=0, size=None, dir=None, mode=None, **kwds):
        """Create a file starting to write at given offset with total size
           of written data given by the size option.

           offset:
               File offset where data will be written to [default: 0]
           size:
               Total number of bytes to write [default: --filesize option]
           dir:
               Create file under this directory
           mode:
               File permissions [default: use default OS permissions]
           pattern:
               Data pattern to write to the file [default: data_pattern default]
           ftype:
               File type to create [default: FTYPE_FILE]
           hole_list:
               List of offsets where each hole is located [default: None]
           hole_size:
               Size of each hole [default: --wsize option]

           Returns the file name created, the file name is also stored
           in the object attribute filename -- attribute absfile is also
           available as the absolute path of the file just created.

           File created is removed at cleanup.
        """
        pattern   = kwds.pop("pattern",   None)
        ftype     = kwds.pop("ftype",     FTYPE_FILE)
        hole_list = kwds.pop("hole_list", None)
        hole_size = kwds.pop("hole_size", self.wsize)

        self.get_filename(dir=dir)
        if size is None:
            size = self.filesize

        if ftype == FTYPE_FILE:
            sfile = None
            self.dprint('DBG3', "Creating file [%s] %d@%d" % (self.absfile, size, offset))
        elif ftype in (FTYPE_SP_OFFSET, FTYPE_SP_ZERO, FTYPE_SP_DEALLOC):
            self.dprint('DBG3', "Creating sparse file [%s] of size %d" % (self.absfile, size))
            sfile = SparseFile(self.absfile, size, hole_list, hole_size)
        else:
            raise Exception("Unknown file type %d" % ftype)

        # Create file
        fd = os.open(self.absfile, os.O_WRONLY|os.O_CREAT|os.O_TRUNC)

        try:
            if ftype == FTYPE_FILE:
                self.write_data(fd, offset, size, pattern)
            elif ftype in [FTYPE_SP_OFFSET, FTYPE_SP_ZERO]:
                for doffset, dsize, dtype in sfile.sparse_data:
                    # Do not write anything to a hole for FTYPE_SP_OFFSET
                    if dtype:
                        self.dprint('DBG4', "    Writing data segment starting at offset %d with length %d" % (doffset, dsize))
                        self.write_data(fd, doffset, dsize, pattern)
                    elif ftype == FTYPE_SP_ZERO:
                        # Write zeros to create the hole
                        self.dprint('DBG4', "    Writing hole segment starting at offset %d with length %d" % (doffset, dsize))
                        self.write_data(fd, doffset, dsize, "\x00")
                if sfile.endhole and ftype == FTYPE_SP_OFFSET:
                    # Extend the file to create the last hole
                    os.ftruncate(fd, size)
            elif ftype == FTYPE_SP_DEALLOC:
                # Create regular file for FTYPE_SP_DEALLOC
                self.dprint('DBG4', "    Writing data segment starting at offset %d with length %d" % (0, size))
                self.write_data(fd, offset, size, pattern)

                for doffset in hole_list:
                    self.dprint('DBG4', "    Create hole starting at offset %d with length %d" % (doffset, hole_size))
                    out = self.libc.fallocate(fd, SR_DEALLOCATE, doffset, hole_size)
                    if out == -1:
                        err = ctypes.get_errno()
                        raise OSError(err, os.strerror(err), self.filename)
        finally:
            os.close(fd)

        if sfile:
            self.sparse_files.append(sfile)
        if mode != None:
            os.chmod(self.absfile, mode)
        return self.filename

    def _reset_files(self):
        """Reset state used in *_files() methods."""
        self.roffset = 0
        self.woffset = 0
        self.rfds = []
        self.wfds = []

    def open_files(self, mode, create=True):
        """Open files according to given mode, the file descriptors are saved
           internally to be used with write_files(), read_files() and
           close_files(). The number of files to open is controlled by
           the command line option '--nfiles'.

           The mode could be either 'r' or 'w' for opening files for reading
           or writing respectively. The open flags for mode 'r' is O_RDONLY
           while for mode 'w' is O_WRONLY|O_CREAT|O_SYNC. The O_SYNC is used
           to avoid the client buffering the written data.
        """
        for i in range(self.nfiles):
            if mode[0] == 'r':
                file = self.abspath(self.files[i])
                self.dprint('DBG3', "Open file for reading: %s" % file)
                fd = os.open(file, os.O_RDONLY)
                self.rfds.append(fd)
                self.lock_type = fcntl.F_RDLCK
            elif mode[0] == 'w':
                if create:
                    self.get_filename()
                    file = self.absfile
                else:
                    file = self.abspath(self.files[i])
                self.dprint('DBG3', "Open file for writing: %s" % file)
                # Open file with O_SYNC to avoid client buffering the write requests
                fd = os.open(file, os.O_WRONLY|os.O_CREAT|os.O_SYNC)
                self.wfds.append(fd)
                self.lock_type = fcntl.F_WRLCK

    def close_files(self):
        """Close all files opened by open_files()."""
        for fd in self.wfds + self.rfds:
            self.dprint('DBG3', "Closing file")
            os.close(fd)
        self._reset_files()

    def write_files(self):
        """Write a block of data (size given by --wsize) to all files opened
           by open_files() for writing.
        """
        for fd in self.wfds:
            self.dprint('DBG4', "Write file %d@%d" % (self.wsize, self.woffset))
            os.write(fd, self.data_pattern(self.woffset, self.wsize))
        self.woffset += self.offset_delta

    def read_files(self):
        """Read a block of data (size given by --rsize) from all files opened
           by open_files() for reading.
        """
        for fd in self.rfds:
            self.dprint('DBG4', "Read file %d@%d" % (self.rsize, self.roffset))
            os.lseek(fd, self.roffset, 0)
            os.read(fd, self.rsize)
        self.roffset += self.offset_delta

    def lock_files(self, lock_type=None, offset=0, length=0):
        """Lock all files opened by open_files()."""
        if lock_type is None:
            lock_type = self.lock_type
        ret = []
        mode_str = 'WRITE' if lock_type == fcntl.F_WRLCK else 'READ'
        lockdata = struct.pack('hhllhh', lock_type, 0, long(offset), long(length), 0, 0)
        for fd in self.rfds + self.wfds:
            try:
                self.dprint('DBG3', "Lock file F_SETLKW (%s)" % mode_str)
                rv = fcntl.fcntl(fd, fcntl.F_SETLKW, lockdata)
                ret.append(rv)
            except Exception, e:
                self.warning("Unable to get lock on file: %r" % e)
        return ret



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
UDP module

Decode UDP layer.
"""
import nfstest_config as c
from baseobj import BaseObj
from packet.utils import ShortHex
from packet.application.dns import DNS
from packet.application.rpc import RPC
from packet.application.ntp4 import NTP
from packet.application.krb5 import KRB5

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.1"

class UDP(BaseObj):
    """UDP object

       Usage:
           from packet.transport.udp import UDP

           x = UDP(pktt)

       Object definition:

       UDP(
           src_port = int,
           dst_port = int,
           length   = int,
           checksum = int,
           data     = string,    # raw data of payload if unable to decode
       )
    """
    # Class attributes
    _attrlist = ("src_port", "dst_port", "length", "checksum", "data")
    _strfmt1  = "UDP {0} -> {1}, len: {2}"
    _strfmt2  = "src port {0} -> dst port {1}, len: {2}, checksum: {3}"

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        unpack = pktt.unpack

        # Decode the UDP layer header
        ulist = unpack.unpack(8, "!HHHH")
        self.src_port = ulist[0]
        self.dst_port = ulist[1]
        self.length   = ulist[2]
        self.checksum = ShortHex(ulist[3])

        pktt.pkt.udp = self
        self._decode_payload(pktt)

    def _decode_payload(self, pktt):
        """Decode UDP payload."""
        if 123 in [self.src_port, self.dst_port]:
            # NTP on port 123
            ntp = NTP(pktt)
            if ntp:
                pktt.pkt.ntp = ntp
            return
        elif 53 in [self.src_port, self.dst_port]:
            # DNS on port 53
            dns = DNS(pktt, proto=17)
            if dns:
                pktt.pkt.dns = dns
            return
        elif 88 in [self.src_port, self.dst_port]:
            # KRB5 on port 88
            krb = KRB5(pktt, proto=17)
            if krb:
                pktt.pkt.krb = krb
            return

        # Get RPC header
        rpc = RPC(pktt, proto=17)

        if rpc:
            # Save RPC layer on packet object
            pktt.pkt.rpc = rpc
            if rpc.type:
                # Remove packet call from the xid map since reply has
                # already been decoded
                pktt._rpc_xid_map.pop(rpc.xid, None)

            # Decode NFS layer
            rpc.decode_payload()



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Host module

Provides a set of tools for running commands on the local host or a remote
host, including a mechanism for running commands in the background.
It provides methods for mounting and unmounting from an NFS server and
a mechanism to simulate a network partition via the use of 'iptables'.
Currently, there is no mechanism to restore the iptables rules to their
original state.
"""
import os
import re
import time
import ctypes
import socket
import subprocess
import nfstest_config as c
from baseobj import BaseObj
from packet.pktt import Pktt

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.3"

class Host(BaseObj):
    """Host object

       Host() -> New Host object

       Usage:
           from nfstest.host import Host

           # Create host object for local host
           x = Host()
           # Create host object for remote host
           y = Host(host='192.168.0.11')

           # Run command to the local host
           x.run_cmd("ls -l")

           # Send command to the remote host and run it as root
           y.run_cmd("ls -l", sudo=True)

           # Run command in the background
           x.run_cmd("tcpdump", sudo=True, wait=False)
           ....
           ....
           # Stop command running in the background
           x.stop_cmd()

           # Mount volume using default options
           x.mount()

           # Unmount volume
           x.umount()

           # Start packet trace
           x.trace_start()

           # Stop packet trace
           x.trace_stop()

           # Open packet trace
           x.trace_open()

           # Enable NFS kernel debug
           x.nfs_debug_enable(nfsdebug='all'):

           # Stop NFS kernel debug
           x.nfs_debug_reset()
    """
    def __init__(self, **kwargs):
        """Constructor

           Initialize object's private data.

           host:
               Hostname or IP address [default: localhost]
           user:
               User to log in to host [default: '']
           server:
               NFS server name or IP address [default: None]
           nfsversion:
               NFS version [default: 4.1]
           proto:
               NFS protocol name [default: 'tcp']
           port:
               NFS server port [default: 2049]
           sec:
               Security flavor [default: 'sys']
           export:
               Exported file system to mount [default: '/']
           mtpoint:
               Mount point [default: '/mnt/t']
           datadir:
               Data directory where files are created [default: '']
           mtopts:
               Mount options [default: 'hard,rsize=4096,wsize=4096']
           interface:
               Network device interface [default: 'eth0']
           nomount:
               Debug option so the server is not actually mounted [default: False]
           tracename:
               Base name for trace files to create [default: 'tracefile']
           trcdelay:
               Seconds to delay before stopping packet trace [default: 0.0]
           tcpdump:
               Tcpdump command [default: '/usr/sbin/tcpdump']
           tbsize:
               Capture buffer size in kB [default: 150000]
           notrace:
               Debug option so a trace is not actually started [default: False]
           rpcdebug:
               Set RPC kernel debug flags and save log messages [default: '']
           nfsdebug:
               Set NFS kernel debug flags and save log messages [default: '']
           dbgname:
               Base name for log messages files to create [default: 'dbgfile']
           messages:
               Location of file for system messages [default: '/var/log/messages']
           tmpdir:
               Temporary directory where trace/debug files are created
               [default: '/tmp']
           iptables:
               Iptables command [default: '/usr/sbin/iptables']
           sudo:
               Sudo command [default: '/usr/bin/sudo']
        """
        # Arguments
        self.host         = kwargs.pop("host",         '')
        self.user         = kwargs.pop("user",         '')
        self.server       = kwargs.pop("server",       '')
        self.nfsversion   = kwargs.pop("nfsversion",   c.NFSTEST_NFSVERSION)
        self.proto        = kwargs.pop("proto",        c.NFSTEST_NFSPROTO)
        self.port         = kwargs.pop("port",         c.NFSTEST_NFSPORT)
        self.sec          = kwargs.pop("sec",          c.NFSTEST_NFSSEC)
        self.export       = kwargs.pop("export",       c.NFSTEST_EXPORT)
        self.mtpoint      = kwargs.pop("mtpoint",      c.NFSTEST_MTPOINT)
        self.datadir      = kwargs.pop("datadir",      '')
        self.mtopts       = kwargs.pop("mtopts",       c.NFSTEST_MTOPTS)
        self.interface    = kwargs.pop("interface",    c.NFSTEST_INTERFACE)
        self.nomount      = kwargs.pop("nomount",      False)
        self.tracename    = kwargs.pop("tracename",    'tracefile')
        self.trcdelay     = kwargs.pop("trcdelay",     0.0)
        self.tcpdump      = kwargs.pop("tcpdump",      c.NFSTEST_TCPDUMP)
        self.tbsize       = kwargs.pop("tbsize",       150000)
        self.notrace      = kwargs.pop("notrace",      False)
        self.rpcdebug     = kwargs.pop("rpcdebug",     '')
        self.nfsdebug     = kwargs.pop("nfsdebug",     '')
        self.dbgname      = kwargs.pop("dbgname",      'dbgfile')
        self.messages     = kwargs.pop("messages",     c.NFSTEST_MESSAGESLOG)
        self.tmpdir       = kwargs.pop("tmpdir",       c.NFSTEST_TMPDIR)
        self.iptables     = kwargs.pop("iptables",     c.NFSTEST_IPTABLES)
        self.sudo         = kwargs.pop("sudo",         c.NFSTEST_SUDO)

        # Initialize object variables
        self.mtdir = self.mtpoint
        self.mounted = False
        self._nfsdebug = False
        self.dbgidx = 1
        self.dbgfile = ''
        self.traceidx = 1
        self.tracefile = ''
        self.tracefiles = []
        self.traceproc = None
        self.process_list = []
        self.process_smap = {}
        self.process_dmap = {}
        self._checkmtpoint = []
        self._invalidmtpoint = []
        self._mtpoint_created = []
        self.need_network_reset = False
        self._localhost = False if len(self.host) > 0 else True
        self.fqdn = socket.getfqdn(self.host)
        ipv6 = self.proto[-1] == '6'
        self.ipaddr = self.get_ip_address(host=self.host, ipv6=ipv6)

        # Load share library - used for functions not exposed in python
        try:
            # Linux
            self.libc = ctypes.CDLL('libc.so.6', use_errno=True)
        except:
            # MacOS
            self.libc = ctypes.CDLL('libc.dylib', use_errno=True)

    def __del__(self):
        """Destructor

           Gracefully unmount volume and reset network.
        """
        self.trace_stop()
        if self.need_network_reset:
            self.network_reset()
        if self.mounted:
            self.umount()
        for mtpoint in self._mtpoint_created:
            try:
                cmd = "rmdir %s" % mtpoint
                self.run_cmd(cmd, sudo=True, dlevel='DBG3', msg="Removing mount point directory: ")
            except:
                pass

    def nfsvers(self, version=None):
        """Return major and minor version for the given NFS version

           version:
               NFS version, default is the object attribute nfsversion
        """
        if version is None:
            version = self.nfsversion
        major = int(version)
        minor = int(round(10*(version-major)))
        return (major, minor)

    def nfsstr(self, version=None):
        """Return the NFS string for the given NFS version

           version:
               NFS version, default is the object attribute nfsversion
        """
        nver, mver = self.nfsvers(version)
        if mver == 0:
            return "NFSv%d" % nver
        else:
            return "NFSv%d.%d" % (nver, mver)

    def abspath(self, filename, dir=None):
        """Return the absolute path for the given file name."""
        bdir = "" if dir is None else "%s/" % dir
        path = "%s/%s%s" % (self.mtdir, bdir, filename)
        return path

    def sudo_cmd(self, cmd):
        """Prefix the SUDO command if effective user is not root."""
        if os.getuid() != 0:
            # Not root -- prefix sudo command
            cmd = self.sudo + ' ' + cmd
        return cmd

    def run_cmd(self, cmd, sudo=False, dlevel='DBG1', msg='', wait=True):
        """Run the command to the remote machine using ssh.
           There is no user authentication, so remote host must allow
           ssh connection without any passwords for the user.
           For a localhost the command is just executed and ssh is not used.

           The object for the process of the command is stored in object
           attribute 'self.process' to be used by methods wait_cmd() and
           stop_cmd(). The standard output of the command is also stored
           in the object attribute 'self.pstdout' while the standard error
           output of the command is stored in 'self.pstderr'.

           cmd:
               Command to execute
           sudo:
               Run command using sudo if option is True
           dlevel:
               Debug level for displaying the command to the user
           msg:
               Prefix this message to the debug message to be displayed
           wait:
               Wait for command to complete before returning

           Return the standard output of the command and the return code or
           exit status is stored in the object attribute 'self.returncode'.
        """
        self.process = None
        self.pstdout = '' 
        self.pstderr = '' 
        self.perror  = '' 
        self.returncode = 0
        if self.user is not None and len(self.user) > 0:
            user = self.user + '@'
        else:
            user = ''

        # Add sudo command if specified
        if sudo:
            cmd = self.sudo_cmd(cmd)

        if not self._localhost:
            cmd = 'ssh -t -t %s%s "%s"' % (user, self.host, cmd.replace('"', '\\"'))

        self.dprint(dlevel, msg + cmd)
        self.process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if not wait:
            self.process_list.append(self.process)
            self.process_dmap[self.process.pid] = dlevel
            if sudo:
                self.process_smap[self.process.pid] = 1
            return
        self.pstdout, self.pstderr = self.process.communicate()
        self.process.wait()
        self.returncode = self.process.returncode
        if self._localhost:
            if self.process.returncode:
                # Error on local command
                self.perror = self.pstderr
                raise Exception(self.pstderr)
        else:
            if self.process.returncode == 255:
                # Error on ssh command
                raise Exception(self.pstderr)
            elif self.process.returncode:
                # Error on command sent
                self.perror = self.pstdout
                raise Exception(self.pstdout)
        return self.pstdout

    def wait_cmd(self, process=None, terminate=False, dlevel=None, msg=''):
        """Wait for command started by run_cmd() to finish.

           process:
               The object for the process of the command to wait for,
               or wait for all commands started by run_cmd() if this
               option is not given
           terminate:
               If True, send a signal to terminate the command or commands
               and then wait for all commands to finish
           dlevel:
               Debug level for displaying the command to the user, default
               is the level given by run_cmd()
           msg:
               Prefix this message to the debug message to be displayed

           Return the exit status of the last command
        """
        if process is None:
            plist = list(self.process_list)
        else:
            plist = [process]

        out = None
        for proc in plist:
            if proc in self.process_list:
                if dlevel is None:
                    _dlevel = self.process_dmap.get(proc.pid, None)
                    if _dlevel is not None:
                        dlevel = _dlevel
                if terminate:
                    if proc.pid in self.process_smap:
                        # This process was started with sudo so kill it with
                        # sudo since terminate() fails to actually kill the
                        # command in RHEL or python 2.6
                        self.run_cmd("kill %d" % proc.pid, sudo=True, dlevel=dlevel, msg=msg)
                        self.process_smap.pop(proc.pid)
                    else:
                        self.dprint(dlevel, msg + "stopping process %d" % proc.pid)
                        proc.terminate()
                else:
                    self.dprint(dlevel, msg + "waiting for process %d" % proc.pid)
                out = proc.wait()
                self.process_list.remove(proc)
        return out

    def stop_cmd(self, process=None, dlevel=None, msg=''):
        """Terminate command started by run_cmd() by calling wait_cmd()
           with the 'terminate' option set to True.

           process:
               The object for the process of the command to terminate,
               or terminate all commands started by run_cmd() if this
               option is not given
           dlevel:
               Debug level for displaying the command to the user, default
               is the level given by run_cmd()
           msg:
               Prefix this message to the debug message to be displayed

           Return the exit status of the last command
        """
        out = self.wait_cmd(process, terminate=True, dlevel=dlevel, msg=msg)
        return out

    def _check_mtpoint(self, mtpoint):
        """Check if mount point exists."""
        if mtpoint in self._checkmtpoint:
            # Run this method once per mount point
            return
        isdir = True
        self._checkmtpoint.append(mtpoint)
        if self._localhost:
            # Locally check if mount point exists and is a directory
            exist = os.path.exists(mtpoint)
            if exist:
                isdir = os.path.isdir(mtpoint)
        else:
            # Remotely check if mount point exists and is a directory
            try:
                cmd = "test -e '%s'" % mtpoint
                self.run_cmd(cmd, dlevel='DBG4', msg="Check if mount point directory exists: ")
            except:
                pass
            exist = not self.returncode
            if exist:
                try:
                    cmd = "test -d '%s'" % mtpoint
                    self.run_cmd(cmd, dlevel='DBG4', msg="Check if mount point is a directory: ")
                except:
                    pass
                isdir = not self.returncode
        if not exist:
            cmd = "mkdir -p %s" % mtpoint
            self.run_cmd(cmd, sudo=True, dlevel='DBG3', msg="Creating mount point directory: ")
            self._mtpoint_created.append(mtpoint)
        elif not isdir:
            self._invalidmtpoint.append(mtpoint)
            raise Exception("Mount point %s is not a directory" % mtpoint)

    def mount(self, **kwargs):
        """Mount the file system on the given mount point.

           server:
               NFS server name or IP address [default: self.server]
           nfsversion:
               NFS version [default: self.nfsversion]
           proto:
               NFS protocol name [default: self.proto]
           port:
               NFS server port [default: self.port]
           sec:
               Security flavor [default: self.sec]
           export:
               Exported file system to mount [default: self.export]
           mtpoint:
               Mount point [default: self.mtpoint]
           datadir:
               Data directory where files are created [default: self.datadir]
           mtopts:
               Mount options [default: self.mtopts]

           Return the mount point.
        """
        # Get options
        server       = kwargs.pop("server",       self.server)
        nfsversion   = kwargs.pop("nfsversion",   self.nfsversion)
        proto        = kwargs.pop("proto",        self.proto)
        port         = kwargs.pop("port",         self.port)
        sec          = kwargs.pop("sec",          self.sec)
        export       = kwargs.pop("export",       self.export)
        mtpoint      = kwargs.pop("mtpoint",      self.mtpoint)
        datadir      = kwargs.pop("datadir",      self.datadir)
        mtopts       = kwargs.pop("mtopts",       self.mtopts)

        # Remove trailing '/' on mount point
        mtpoint = mtpoint.rstrip("/")

        if len(datadir):
            self.mtdir = os.path.join(mtpoint, datadir)
        else:
            self.mtdir = mtpoint

        self._check_mtpoint(mtpoint)
        if self.nomount or mtpoint in self._invalidmtpoint:
            return

        if len(export) > 1:
            # Remove trailing '/' on export path if is not the root directory
            export = export.rstrip("/")

        # Using the proper version of NFS
        mt_list = []
        nver, mver = self.nfsvers(nfsversion)
        if mver > 0:
            mt_list.append("vers=%d.%d" % (nver, mver))
        else:
            mt_list.append("vers=%d" % nver)

        if port != 2049:
            mt_list.append("port=%d" % port)

        mt_list.extend(["proto=%s"%proto, "sec=%s"%sec, mtopts])
        mtopts = ",".join(mt_list)

        # Mount command
        cmd = "mount -o %s %s:%s %s" % (mtopts, server, export, mtpoint)
        self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="Mount volume: ")

        self.mounted = True
        self.mtpoint = mtpoint

        # Create data directory if it does not exist
        if self._localhost:
            if not os.path.exists(self.mtdir):
                os.mkdir(self.mtdir, 0777)
        else:
            try:
                cmd = "test -e '%s'" % self.mtdir
                self.run_cmd(cmd, dlevel='DBG4', msg="Check if data directory exists: ")
            except:
                pass
            if self.returncode:
                cmd = "mkdir -m 0777 -p %s" % self.mtdir
                self.run_cmd(cmd, dlevel='DBG3', msg="Creating data directory: ")

        # Return the mount point
        return mtpoint

    def umount(self):
        """Unmount the file system."""
        if self.nomount:
            return

        self._check_mtpoint(self.mtpoint)
        if self.mtpoint in self._invalidmtpoint:
            return

        self.dprint('DBG3', "Sync all buffers to disk")
        self.libc.sync()

        # Try to umount 5 times
        cmd = "umount -f %s" % self.mtpoint
        for i in range(5):
            time.sleep(1)
            try:
                self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="Unmount volume: ")
            except:
                pass

            if self.returncode == 0 or re.search('not (mounted|found)', self.perror):
                # Unmount succeeded or directory not mounted
                self.mounted = False
                break
            self.dprint('DBG2', self.perror)

    def trace_start(self, tracefile=None, interface=None, capsize=None, clients=None):
        """Start trace on interface given

           tracefile:
               Name of trace file to create, default is a unique name
               created in the temporary directory using self.tracename as the
               base name.
           capsize:
               Use the -C option of tcpdump to split the trace files every
               1000000*capsize bytes. See documentation for tcpdump for more
               information
           clients:
               List of Host() objects to monitor

           Return the name of the trace file created.
        """
        self.trace_stop()
        if tracefile:
            self.tracefile = tracefile
        else:
            self.tracefile = "%s/%s_%d.cap" % (self.tmpdir, self.tracename, self.traceidx)
            self.traceidx += 1
        if not self.notrace:
            if len(self.nfsdebug) or len(self.rpcdebug):
                self.nfs_debug_enable()
            self.tracefiles.append(self.tracefile)

            if clients is None:
                clients = self.clients

            if interface is None:
                interface = self.interface

            opts = ""
            if interface is not None:
                opts += " -i %s" % interface

            if capsize:
                opts += " -C %d" % capsize

            hosts = self.ipaddr
            for cobj in clients:
                hosts += " or %s" % cobj.ipaddr

            cmd = "%s%s -n -B %d -s 0 -w %s host %s" % (self.tcpdump, opts, self.tbsize, self.tracefile, hosts)
            self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="Trace start: ", wait=False)
            self.traceproc = self.process

            # Make sure tcpdump has started
            out = self.traceproc.stderr.readline()
            if not re.search('listening on', out):
                time.sleep(1)
                if self.process.poll() is not None:
                    raise Exception(out)
        return self.tracefile

    def trace_stop(self):
        """Stop the trace started by trace_start()."""
        try:
            if self.traceproc:
                self.dprint('DBG2', "Trace stop")
                time.sleep(self.trcdelay)
                self.run_cmd("killall tcpdump", sudo=True, dlevel='DBG2')
                self.stop_cmd(self.traceproc)
                self.traceproc = None
            if not self.notrace and self._nfsdebug:
                self.nfs_debug_reset()
        except:
            return

    def trace_open(self, tracefile=None, **kwargs):
        """Open the trace file given or the trace file started by trace_start().

           All extra options are passed directly to the packet trace object.

           Return the packet trace object created, the packet trace object
           is also stored in the object attribute pktt.
        """
        if tracefile is None:
            tracefile = self.tracefile
        self.dprint('DBG1', "trace_open [%s]" % tracefile)
        self.pktt = Pktt(tracefile, **kwargs)
        return self.pktt

    def nfs_debug_enable(self, **kwargs):
        """Enable NFS debug messages.

           rpcdebug:
               Set RPC kernel debug flags and save log messages [default: self.rpcdebug]
           nfsdebug:
               Set NFS kernel debug flags and save log messages [default: self.nfsdebug]
           dbgfile:
               Name of log messages file to create, default is a unique name
               created in the temporary directory using self.dbgname as the
               base name.
        """
        modmsgs = {
            'nfs': kwargs.pop('nfsdebug', self.nfsdebug),
            'rpc': kwargs.pop('rpcdebug', self.rpcdebug),
        }
        dbgfile = kwargs.pop('dbgfile', None)
        if dbgfile is not None:
            self.dbgfile = dbgfile
        else:
            self.dbgfile = "%s/%s_%d.msg" % (self.tmpdir, self.dbgname, self.dbgidx)
            self.dbgidx += 1

        if modmsgs['nfs'] is None and modmsgs['rpc'] is None:
            return

        if os.path.exists(self.messages):
            fstat = os.stat(self.messages)
            self.dbgoffset = fstat.st_size
            self.dbgmode = fstat.st_mode & 0777
            for mod in modmsgs.keys():
                if len(modmsgs[mod]):
                    self._nfsdebug = True
                    cmd = "rpcdebug -v -m %s -s %s" % (mod, modmsgs[mod])
                    self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="NFS debug enable: ")

    def nfs_debug_reset(self):
        """Reset NFS debug messages."""
        for mod in ('nfs', 'rpc'):
            try:
                cmd = "rpcdebug -v -m %s -c" % mod
                self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="NFS debug reset: ")
            except:
                pass

        if self.dbgoffset != None:
            try:
                fd = None
                fdw = None
                os.system(self.sudo_cmd("chmod %o %s" % (self.dbgmode|0444, self.messages)))
                self.dprint('DBG2', "Creating log messages file [%s]" % self.dbgfile)
                fdw = open(self.dbgfile, "w")
                fd = open(self.messages, "r")
                fd.seek(self.dbgoffset)
                while True:
                    data = fd.read(self.rsize)
                    if len(data) == 0:
                        break
                    fdw.write(data)
            except Exception as e:
                raise
            finally:
                if fd:
                    fd.close()
                if fdw:
                    fdw.close()
                os.system(self.sudo_cmd("chmod %o %s" % (self.dbgmode, self.messages)))

    def network_drop(self, ipaddr, port):
        """Simulate a network drop by dropping all tcp packets going to the
           given ipaddr and port using the iptables commands.
        """
        self.need_network_reset = True
        cmd = "%s -A OUTPUT -p tcp -d %s --dport %d -j DROP" % (self.iptables, ipaddr, port)
        self.run_cmd(cmd, sudo=True, dlevel='DBG6', msg="Network drop: ")

    def network_reset(self):
        """Reset the network by flushing all the chains in the table using
           the iptables command.
        """
        try:
            cmd = "%s --flush" % self.iptables
            self.run_cmd(cmd, sudo=True, dlevel='DBG6', msg="Network reset: ")
        except:
            self.dprint('DBG6', "Network reset error <%s>" % self.perror)

        try:
            cmd = "%s --delete-chain" % self.iptables
            self.run_cmd(cmd, sudo=True, dlevel='DBG6', msg="Network reset: ")
        except:
            self.dprint('DBG6', "Network reset error <%s>" % self.perror)

    def get_route(self, ipaddr):
        """Get routing information for destination IP address
           Returns a tuple: (gateway, device name, src IP address)
        """
        try:
            cmd = "ip route get %s" % ipaddr
            out = self.run_cmd(cmd, dlevel='DBG1', msg="Get routing info: ")
            regex = re.search(r"(\svia\s+(\S+))?\sdev\s+(\S+).*\ssrc\s+(\S+)", out)
            if regex:
                return regex.groups()[1:]
        except:
            self.dprint('DBG7', self.perror)
        return (None, None, None)

    @staticmethod
    def get_ip_address(host='', ipv6=False):
        """Get IP address associated with the given host name.
           This could be run as an instance or class method.
        """
        lstr = ""
        ipstr = "v6" if ipv6 else "v4"
        family = socket.AF_INET6 if ipv6 else socket.AF_INET
        if len(host) == 0:
            lstr = "local "
            host = socket.gethostname()

        try:
            infolist = socket.getaddrinfo(host, 2049, 0, 0, socket.SOL_TCP)
        except Exception as e:
            infolist = []
        for info in infolist:
            # Ignore loopback addresses
            if info[0] == family and info[4][0] not in ('127.0.0.1', '::1'):
                return info[4][0]
        raise Exception("Unable to get IP%s address for %shost '%s'" % (ipstr, lstr, host))
